import importlib.resources as importlib_resources


def get_lib_resource_path(package, resource):
    """Get the path to a resource in a package.

    Parameters:
    package: str
        The name of the package containing the resource.
    resource: str
        The name of the resource.

    Returns:
    resource_path: pathlib.PosixPath
        Path to the resource.
    """
    resource_path = importlib_resources.path(package, resource)
    if hasattr(resource_path, 'args'):
        resource_path, = resource_path.args
    return resource_path

def path_to_workflows():
    """Get the path to the workflows directory.

    Returns:
    workflows_path: pathlib.PosixPath
        Path to the workflows directory.
    """
    workflows_path = get_lib_resource_path('beast_pype', 'workflows')
    return workflows_path

def path_to_report_templates():
    """Get the path to the report templates directory.

    Returns:
    reports_path: pathlib.PosixPath
        Path to the report templates directory.
    """
    reports_path = get_lib_resource_path('beast_pype', 'report_templates')
    return reports_path

def path_to_workflow_modules():
    """Get the path to the workflow modules directory.

    Returns:
    workflow_modules_path: pathlib.PosixPath
        Path to the workflow modules directory.
    """
    workflow_modules_path = get_lib_resource_path('beast_pype', 'workflow_modules')
    return workflow_modules_path


import papermill as pm
import nbformat as nbf

def change_notebook_kernel(input_path, output_path, kernel):
    """
    Change a Jupyter Notebook's kernel_name.

    Parameters
    ----------
    input_path: str
        Path to the input notebook.
    output_path: str
        Path to the output notebook.
    kernel: str
        Name of the kernel_name to change to.
    """
    notebook = nbf.read(input_path, as_version=4)
    notebook['metadata']['kernelspec']['name'] = kernel
    notebook['metadata']['kernelspec']['display_name'] = kernel
    nbf.write(notebook, output_path)

def execute_notebook(input_path,
                     output_path,
                     kernel_name=None,
                     **kwargs
                     ):
    """
    Execute a Jupyter Notebook using a specific Jupyter kernel_name.

    Notes
    ------
    This function is wrapper for papermill.execute, as at papermill v 2.6.0 the
    `kernal_name` argument for that function is not working.

    Parameters
    ----------
    input_path: str
        Path to input notebook.
    output_path: str
        Path to save executed notebook.
    kernel_name: str, optional
        Name of the kernel_name to execute.
    kwargs:
        Additional arguments to pass to papermill.execute.
    """
    if kernel_name is not None:
        change_notebook_kernel(input_path, output_path, kernel_name)
        pm.execute_notebook(input_path=output_path,
                            output_path=output_path,
                            kernel_name=kernel_name, **kwargs)
    else:
        pm.execute_notebook(input_path=input_path, output_path=output_path, **kwargs)


from copy import deepcopy
from pandas.tseries.offsets import DateOffset

def gen_partition_dates(partition_freq, date_of_youngest_tip, voi_youngest_oldest_tip, voi_strains, dr_strain):
    """

    Parameters
    ----------
    partition_freq
    date_of_youngest_tip
    voi_youngest_oldest_tip
    voi_strains
    dr_strain

    Returns
    -------

    """
    voi_partition_dates = []
    offsets = {partition_freq['unit']: partition_freq['every']}
    date_to_check = date_of_youngest_tip - DateOffset(**offsets)
    if not date_to_check > voi_youngest_oldest_tip:
        raise ValueError(
            f'Change dates dictionary for partition_freq makes no sense.\n' +
            f"First partition date generated is not after the most recent tip date out of oldest tips for each VOI lineage  ({voi_youngest_oldest_tip.strftime('%Y-%m-%d')}).\n" +
            'Suggest using smaller "units" or "every" values or pushing "end" date back.'
        )
    while date_to_check > voi_youngest_oldest_tip:
        date_to_append = date_to_check
        voi_partition_dates.append(date_to_append.strftime('%Y-%m-%d'))
        date_to_check= date_to_append - DateOffset(**offsets)

    dr_partition_dates = deepcopy(voi_partition_dates)
    if 'dr_extra_offset' in partition_freq:
        dr_extra_offset = {partition_freq['dr_extra_offset']['unit']: partition_freq['dr_extra_offset']['amount']}
    else:
        dr_extra_offset = offsets
    date_to_append = date_to_append - DateOffset(**dr_extra_offset )
    dr_partition_dates.append(date_to_append.strftime('%Y-%m-%d'))
    partition_dates = {f"VOI_{voi_strain}": voi_partition_dates for voi_strain in voi_strains}
    partition_dates[f'DR_{dr_strain}'] = dr_partition_dates
    return partition_dates


import pandas as pd
from random import sample


def split_df_by_weeks(df, date_field='date'):
    """ 
    Splits a dataframe into chunks of a week based on Monday being the first day of the week.
    """
    return [g for n, g in df.groupby(pd.Grouper(key=date_field, freq='W'))]
    
    
def df_weekly_sampler(df, n, draws=1, date_field='date'):
    """
    Draw a random xml_set from a dataframe spread evenly across weekly chunks.
    
    If a week has fewer rows than the xml_set needed then the remainder is spread
     across all other rows.
    Any remainder after this is distributed by drawing an extra row from other
    weeks selected at random.
    
    Parameters
    ----------
    df : pandas.DataFrame
        Pandas DataFrame from which the random samples are drawn.
    n : int
        Total number of samples to draw.
    draws : int
        Number of draws of size N you wish to make. Replacement occurs between draws.
    date_field : str, default 'date'
        Field of date variable to base chopping up selection into weeks.

    Returns
    -------
    draws_data : pandas.DataFrame or Mast of pandas.DataFrames
        If draws > 1 a list of pandas.DataFrames is returned, otherwise a single dataframe.
    
    """
    if len(df) < n:
        raise ValueError('Dataframe has less rows than the xml_set size selected.')
    if len(df) == n:
        raise ValueError('Dataframe has the same number of rows as the xml_set size selected.')

    week_split_dfs = split_df_by_weeks(df, date_field=date_field)
    number_of_weeks = len(week_split_dfs)
    modulus = n % number_of_weeks
    typical_weekly_sample = int(n/number_of_weeks)

    available_per_week = [len(week_df) for week_df in week_split_dfs]

    less_than_equal_typical_sample = [available <= typical_weekly_sample for available in available_per_week]
    less_than_needed = typical_weekly_sample - pd.Series(available_per_week)[less_than_equal_typical_sample]
    extras_needed = less_than_needed.sum()+modulus

    if extras_needed > 0:
        n_weeks_more_than_needed = less_than_equal_typical_sample.count(False)
        extras_needed_per_available = int(extras_needed / n_weeks_more_than_needed)
        typical_weekly_sample += extras_needed_per_available
        less_than_equal_typical_sample = [available <= typical_weekly_sample for available in available_per_week]
        index_greater_than_typical_sample = [
            week_i for week_i, available in
            enumerate(available_per_week)
            if available > typical_weekly_sample
        ]
        n_weeks_more_than_needed = len(index_greater_than_typical_sample)
        final_modulus = extras_needed % n_weeks_more_than_needed
        weeks_to_add_one_extra = sample(index_greater_than_typical_sample, final_modulus)
    else:
        weeks_to_add_one_extra = []

    draws_dfs = []
    for draw in range(draws):
        sampled_data = []
        for week_i, weeK_df in enumerate(week_split_dfs):
            if less_than_equal_typical_sample[week_i]:
                sampled_data.append(weeK_df)
            else:
                this_weeks_sample = typical_weekly_sample + int(week_i in weeks_to_add_one_extra)
                sampled_data.append(weeK_df.xml_set(this_weeks_sample))
        sampled_data = pd.concat(sampled_data)
        draws_dfs.append(sampled_data)
    
    if draw == 0:
        return draws_dfs[0]
    else:
        return draws_dfs


from warnings import warn
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from arviz import hdi
from beast_pype.date_utilities import decimal_to_date, date_to_decimal
from beast_pype.fig_utils import year_decimal_to_date_tick_labels
from scipy.interpolate import interp1d
from copy import deepcopy
import seaborn as sns
import re

# stop annoying panadas warnings
pd.options.mode.chained_assignment = None



def read_log_file(file_path,
                  cut_to_first=None,
                  remove_0_first_sampling_prop=True,
                  youngest_tip_date=None,
                  convert_become_uninfectious_rate=False,
                  ):
    """Read a log file into a pandas DataFrame.

    Parameters
    ----------------
    file_path: str
        Path to the log file.
    cut_to_first : int, default None
        Remove Samples/links over this number.
    remove_0_first_sampling_prop : bool, default True
        If all the values in the first column starting with 'samplingProportion' are equal
        to 0, remove that column.
    youngest_tip_date : datetime, default None
        Youngest tip date. If supplied an attempt are made to:
         * determine the treeheight column and convert to TMRCA (through subtraction from youngest_tip_date).
         * determine the BDSKY Origin column and convert to Origin (through subtraction from youngest_tip_date).
    convert_become_uninfectious_rate:   bool, default False
        If true column 'becomeUninfectiousRate_BDSKY_Serial' will be used to
         calculate columns 'Rate of Becoming Uninfectious (per day)' and
          'Infection period (per day)'.

    Returns
    -----------
    Pandas.DataFrame
    """
    trace_df = pd.read_table(file_path, sep='\t', comment='#')
    if cut_to_first is not None:
        trace_df = trace_df[trace_df['Sample'] <= cut_to_first]
    if remove_0_first_sampling_prop:
        sampling_prop_columns = [column for column in trace_df.columns if column.startswith('samplingProportion')]
        if len(sampling_prop_columns) > 0 and (trace_df[sampling_prop_columns[0]] == 0).all():
            trace_df.drop(sampling_prop_columns[0], axis=1, inplace=True)
    if youngest_tip_date is not None:
        if not isinstance(youngest_tip_date, (int, float)):
            youngest_tip_date = date_to_decimal(youngest_tip_date)
        possible_origins = [column
                            for column in trace_df.columns
                            if 'origin' in column.lower()]
        if len(possible_origins) >1:
            warn(f'More than one column contains the substring "origin" (after conversion to lowercase) in trace at {file_path}.' +
                 'Therefore conversion to "Origin" column by subtraction from youngest tip date skipped.')
        if len(possible_origins) ==1:
            column_starting_origin = possible_origins[0]
            trace_df['Origin'] = youngest_tip_date - trace_df[column_starting_origin]
            # trace_df.drop(columns=column_starting_origin, inplace=True)
            # If this column is deleted it screws up the _gridded_skyline function
            # needed in plotting smoothed skylines. I do not have time to debug the
            # resulting break.

        possible_tree_heights = [column
                                    for column in trace_df.columns
                                    if 'treeheight' in re.sub(r'[^a-z]', '', column.lower())]
        if len(possible_tree_heights) >1:
            warn(f'More than one column contains the substring "treeheight" (after conversion to lowercase and removal of non-letters) in trace at {file_path}.' +
                 'Therefore conversion to "TMRCA" column by subtraction from youngest tip date skipped.')
        elif len(possible_tree_heights) ==1:
            tree_height = possible_tree_heights[0]
            trace_df['TMRCA'] = youngest_tip_date - trace_df[tree_height]
            # trace_df.drop(columns=tree_height, inplace=True)
            # Deleting this column caused a few bugs. Don't have time to sortout.
        else:
            warn(
                f'No column contains the substring "treeheight" (after conversion to lowercase and removal of non-letters) in trace at {file_path}.' +
                'Therefore conversion to "TMRCA" column by subtraction from youngest tip date skipped.')
        if convert_become_uninfectious_rate:
            trace_df['Rate of Becoming Uninfectious (per day)'] = trace_df['becomeUninfectiousRate_BDSKY_Serial'] / 365
            trace_df['Mean Infection period (days)'] = 1 / trace_df['Rate of Becoming Uninfectious (per day)']
    return trace_df


def read_xml_set_logs_for_plotting(file_path_dict,
                                   xml_set_label='xml_set',
                                   remove_0_first_sampling_prop=True,
                                   convert_become_uninfectious_rate=False,
                                   youngest_tip_dates_dict=None):
    """
    Read xml set log files into a pandas DataFrame and a melted dataframe.

    Parameters
    ----------
    file_path_dict: dict {str: str}
        Dictionary of paths to log files. Keys are strain names.
    xml_set_label: str default 'xml_set'
        Label for xml_set divisions e.g. 'country' or 'strain'.
    remove_0_first_sampling_prop : bool, default True
        If all the values in the first column starting with 'samplingProportion' are equal
        to 0, remove that column.
    convert_become_uninfectious_rate:   bool, default False
        If true column 'becomeUninfectiousRate_BDSKY_Serial' will be used to
         calculate columns 'Rate of Becoming Uninfectious (per day)' and
          'Infection period (per day)'.
    youngest_tip_dates_dict: dict {str: float}
        Dictionary of youngest tip dates. Keys are xml_set names.


    Returns
    -------

    """
    dfs_list = []
    for xml_set, file_path in file_path_dict.items():
        if youngest_tip_dates_dict is not None:
            youngest_tip = youngest_tip_dates_dict[xml_set]
        else:
            youngest_tip = youngest_tip_dates_dict
        temp_df = read_log_file(file_path,
                                remove_0_first_sampling_prop=remove_0_first_sampling_prop,
                                youngest_tip_date=youngest_tip,
                                convert_become_uninfectious_rate=convert_become_uninfectious_rate)
        temp_df.insert(loc=0, column=xml_set_label, value=xml_set)
        dfs_list.append(temp_df)

    df = pd.concat(dfs_list)

    id_vars = [xml_set_label]
    value_vars = [col for col in df.columns if col not in id_vars]
    df_melt = df.melt(id_vars=id_vars,
                      value_vars=value_vars,
                      value_name='Estimate')
    return df, df_melt


def percentile_5th(g):
    return np.percentile(g, 5)

def percentile_95th(g):
    return np.percentile(g, 95)

def percentile_pivot(df, column, xml_set_label='xml_set'):
    """
    Create a percentile based pivot tale.
    
    """
    df_to_return = pd.pivot_table(df, values=[column], index=[xml_set_label],
                                  aggfunc= [percentile_5th, np.median, percentile_95th])
    if column == 'Origin':
        df_to_return = df_to_return.map(decimal_to_date).map(lambda x: x.strftime('%Y-%m-%d'))
    df_to_return.index = df_to_return.index.str.split('_', expand=True)
    df_to_return.reset_index(inplace=True)
    df_to_return.columns = ['Type of Strain', 'Strain', 'Sample Size','5th Percentile', 'Median (50th Percentile)', '95th Percentile']
    return df_to_return


def hdi_columns_starting_with(df, starting_with, hdi_prob=0.95):
    """
    Calculate HDI for columns starting with a given string.

    Parameters
    ----------
    df: pd.DataFrame
    starting_with: str
        C
    hdi_prob: float, default=0.95
        HDI probability to use.

    Returns
    -------
    pd.DataFrame

    """
    records = []
    cols_starting_with = [col for col in df.columns if col.startswith(starting_with)]
    for column in cols_starting_with:
        selected_value = df[column].to_numpy()
        lower_interval, upper_interval = hdi(selected_value, hdi_prob=hdi_prob)
        median_val = np.median(selected_value)
        records.append({'Parameter': column,
                        f'Lower {str(hdi_prob)} HDI': lower_interval,
                        'Median': median_val,
                        f'Upper {str(hdi_prob)} HDI': upper_interval})
    return pd.DataFrame.from_records(records)

def hdi_pivot(df, column, xml_set_label='xml_set', hdi_prob=0.95):
    """
    Create a hdi based pivot tale.

    Parameters
    ----------
    df: pd.DataFrame
        Data Frame to be pivoted.
    column: str
        Column to calculate HDI for.
    xml_set_label: str, optional
        Label for XML set, if not provided 'xml_set' is used.
    hdi_prob: float, default=0.95
        HDI probability to use.

    Returns
    -------
    pd.DataFrame
    """
    records = []
    for selection in df[xml_set_label].unique():
        selected_value = df[column][df[xml_set_label]==selection].to_numpy()
        lower_interval, upper_interval = hdi(selected_value, hdi_prob=hdi_prob)
        median_val = np.median(selected_value)
        records.append(
            {xml_set_label: selection,
             f'Lower {str(hdi_prob)} HDI': lower_interval,
             'Median':median_val,
             f'Upper {str(hdi_prob)} HDI': upper_interval
             })
    
    return pd.DataFrame.from_records(records)

def summary_stats_and_plot(df,
                           x,
                           y,
                           convert_plot_to_seconds=False,
                           include_grid=True,
                           violinplot_kwargs={},
                           boxplot_kwargs={}):
    """
    Calculate summary statistics and plot as box violin plots.

    Parameters
    ----------
    df: pandas.DataFrame
        Melted DataFrame for use in Seaborne.
    x: str
        Column to plot on x-axis.
    y: str
        Column to plot on y-axis.
    convert_plot_to_seconds: bool, default=False
        Whether to convert plot to seconds.
    include_grid: bool, default=True
        Whether to include grid.
    violinplot_kwargs: dict
        Keyword arguments to pass to seaborn.violinplot.
    boxplot_kwargs: dict
        Keyword arguments to pass to seaborn.boxplot.

    Returns
    -------
    fig: Seaborn plot
        Seaborn box violin plots.
    y_df: pandas.DataFrame
        DataFrame with columns specified in y and x.
        If convert_plot_to_second ==True an additional column where y is converted
        to seconds is included.
    summary_stats: pandas.DataFrame
        Summary statistics.
    """
    df = deepcopy(df)
    if convert_plot_to_seconds:
        df[f'{y} seconds'] = df[y].dt.total_seconds()
        columns = [x, f'{y} seconds', y]
        plot_y = f'{y} seconds'
    else:
        columns = [x, y]
        plot_y = y

    y_df = df[columns]
    summary_stats = y_df.groupby(x).describe().transpose()
    fig = plot_box_violin(df=y_df,
                          x=x,
                          y=plot_y,
                          include_grid=include_grid,
                          violinplot_kwargs=violinplot_kwargs,
                          boxplot_kwargs=boxplot_kwargs
                          )
    return fig, y_df, summary_stats



def plot_box_violin(df, x, y, include_grid=True, violinplot_kwargs={}, boxplot_kwargs={}):
    """
    Plot box violin plots.

    Parameters
    ----------
    df: pandas.DataFrame
        Melted DataFrame for use in Seaborne.
    x: str
        Column to plot on x-axis.
    y: str
        Column to plot on y-axis.
    include_grid: bool, default=True
        Whether to include grid.
    violinplot_kwargs: dict
        Keyword arguments to pass to seaborn.violinplot.
    boxplot_kwargs: dict
        Keyword arguments to pass to seaborn.boxplot.

    Returns
    -------
    Seaborn box violin plots.
    """
    fig = sns.violinplot(data=df, y=y, x=x, inner=None, hue=x, **violinplot_kwargs)
    sns.boxplot(data=df, y=y, x=x, ax=fig, color='grey', width=0.3, **boxplot_kwargs)
    fig.set(ylabel=y)
    fig.set_xticklabels(fig.get_xticklabels(), rotation=90)
    plt.tight_layout()
    if include_grid:
        plt.grid(linestyle='-.')
    return fig


def plot_comparative_box_violin(df_melted, parameter, xml_set_label='xml_set', prior_draws=None,
                                include_grid=True, violinplot_kwargs={}, boxplot_kwargs={}):
    """
    Plot box violin plots comparing xml_sets.

    Parameters
    ----------
    df_melted : pandas.DataFrame
        Melted DataFrame for use in Seaborne.  Contains columns 'xml_set', 'variable' and
        'Estimate'.
    parameter   :   str
        Parameter name.
    xml_set_label: str, optional
        Label for x-axis, if not provided this will be 'xml_set'.
    prior_draws :   numpy.ndarray,default None
        If given included in plot as 'Draws from Prior'
    include_grid:   bool, default=True
        Whether to include grid in violin plot.
    violinplot_kwargs: dict
        Keyword arguments to pass to seaborn.violinplot.
    boxplot_kwargs: dict
        Keyword arguments to pass to seaborn.boxplot.

    Returns
    -------
    Seaborn box violin plots.
    """
    df_melted = df_melted[df_melted['variable'] == parameter]
    if prior_draws is not None:
        temp_data = {xml_set_label: 'Draws from Prior',
                     'Sample': None,
                     'variable': parameter,
                     'Estimate': prior_draws}
        temp_df = pd.DataFrame(temp_data)
        df_melted = pd.concat([df_melted, temp_df])

    fig = plot_box_violin(df=df_melted,
                          x=xml_set_label,
                          y='Estimate',
                          include_grid=include_grid,
                          violinplot_kwargs=violinplot_kwargs,
                          boxplot_kwargs=boxplot_kwargs)
    return fig


def _set_dates_skyline_plotting_df(log_df, parameter_start, style, partition_year_decimals, hdi_prob=0.95):
    """Generate a DataFrame for plotting HDI values from a skyline model log.

    Parameters
    ---------------
    log_df: pd.DataFrame
        BDSKY log dataframe.
    hdi_prob: float
        Highest density interval to be used.


    Returns
    -----------
    parameter_plotting_df: pd.DataFrame
        HDI and median R_t values with deci dates (year fractions) and dates.
    """
    parameter_df = log_df.loc[:, log_df.columns.str.startswith(parameter_start)]
    parameter_df.dropna(axis=1, how='all', inplace=True)
    parameter_hdi = pd.DataFrame({label: hdi(series.to_numpy(), hdi_prob=hdi_prob) for label, series in parameter_df.items()})
    parameter_median = parameter_df.median()
    parameter_values = pd.concat([parameter_hdi.iloc[0, :], parameter_median, parameter_hdi.iloc[-1, :]], axis=1)
    parameter_values.columns = ['lower', 'median', 'upper']
    if parameter_df.columns[0].startswith('sampling') and not parameter_df.columns[0].endswith('.1'):
        # This only happens if the firs sampling proportion has been dropped in read_log_file
        # for being fixed at 0.
        year_decimals = pd.Series(np.sort(partition_year_decimals))
    else:
        start = log_df['Origin'].median()
        year_decimals = pd.Series(np.sort(np.append(partition_year_decimals, start)))
    mid_points = [np.mean([i, j])
                  for i, j in zip(year_decimals[:-1], year_decimals[1:])]
    if style.startswith('skyline'):
        parameter_plotting_df = pd.DataFrame({
            'year_decimal': year_decimals.repeat(2)[1:-1].to_list(),
            'lower': parameter_values['lower'].repeat(2).to_list(),
            'median': parameter_values['median'].repeat(2).to_list(),
            'upper': parameter_values['upper'].repeat(2).to_list()
        })
    elif style.startswith('smooth spline'):
        kind = 'quadratic'
        lower_spline = interp1d(mid_points, parameter_values['lower'], kind=kind)
        median_spline = interp1d(mid_points, parameter_values['median'], kind=kind)
        upper_spline = interp1d(mid_points, parameter_values['upper'], kind=kind)
        new_x_vals = np.linspace(mid_points[0], mid_points[-1], 1000)
        parameter_plotting_df = pd.DataFrame({
            'year_decimal': new_x_vals,
            'lower': lower_spline(new_x_vals),
            'median': median_spline(new_x_vals),
            'upper': upper_spline(new_x_vals)
        })
    else:
        raise ValueError('style start with "skyline" or "smooth spline".')
    if style.endswith('with mid-points'):
        mid_point_df = pd.DataFrame({
            'year_decimal': mid_points,
            'lower': parameter_values['lower'],
            'median': parameter_values['median'],
            'upper': parameter_values['upper']
        })
        return parameter_plotting_df, mid_point_df
    else:
        return parameter_plotting_df


def _gridded_skyline(trace_df,
                     youngest_tip,
                     parameter_start='reproductiveNumber',
                     grid_size=100):
    """
    Create gridded skyline parameters from Birth-Death Skyline trace over time.

    Parameters
    ----------
    trace_df : DataFrame
        Data frame of BEAST2 MCMC runs.
    youngest_tip : float
        Year decimal of youngest tip.
    parameter_start : str
        Starting string of parameters.
    grid_size: int, default 100
        Grid size for smoothing skyline interpolation.


    Returns
    -------
    Pandas.Dataframe

    Notes
    --------
    Skyline griding interpolation creation adapted from the R function gridSkyline
    from https://github.com/laduplessis/bdskytools/tree/master
    """
    parameter_subset = trace_df.loc[:, trace_df.columns.str.startswith(parameter_start)]
    origin_column = [column for column in trace_df.columns if column.startswith('origin')][0]
    origins = trace_df[origin_column]
    origin_med = origins.median()
    grid_times = np.linspace(start=0, stop=origin_med, num=grid_size)
    links, sky_grids = parameter_subset.shape
    sky_matrix = parameter_subset.to_numpy()

    # skyline_gridded creation adapted from the R function gridSkyline from
    # https://github.com/laduplessis/bdskytools/tree/master
    skyline_gridded = []
    for i in range(links):
        skyline_indices = np.maximum(1,
                                     sky_grids - np.floor(grid_times / origins[i] * sky_grids)
                                     ).astype(int) - 1
        skyline_gridded.append(sky_matrix[i, skyline_indices])

    skyline_gridded = pd.DataFrame(skyline_gridded, columns=grid_times)
    skyline_gridded_hpd = skyline_gridded.apply(lambda x: hdi(x.to_numpy(),
                                                              hdi_prob=0.95), axis=0, result_type='expand')
    skyline_gridded_hpd = skyline_gridded_hpd.transpose()
    skyline_gridded_hpd.columns = ['lower', 'upper']
    skyline_gridded_hpd.insert(1, 'median', skyline_gridded.median())
    skyline_gridded_hpd['year_decimal'] = youngest_tip - grid_times
    return skyline_gridded_hpd

def _plot_df_to_hdi_df(plot_df):
    hdi_est_df = plot_df[plot_df.index % 2 == 0]
    end_df = plot_df[plot_df.index % 2 != 0]
    hdi_est_df['Start of Period'] = hdi_est_df.year_decimal.map(decimal_to_date).map(
        lambda x: x.strftime('%Y-%m-%d'))
    hdi_est_df['End of Period'] = end_df.year_decimal.map(decimal_to_date).map(
        lambda x: x.strftime('%Y-%m-%d')).to_list()
    return hdi_est_df[['Start of Period', 'End of Period', 'lower', 'median', 'upper']]
    
    

def plot_skyline(traces_df,
                 youngest_tip_year_decimal,
                 parameter_start='reproductiveNumber',
                 x_tick_freq='automatic',
                 style='skyline',
                 y_label='$R_t$',
                 palette=sns.color_palette(),
                 include_grid=True,
                 grid_size=100,
                 partition_year_decimals=None,
                 xml_set_label='xml_set'
                 ):
    """
    Plot parameters skylines with set dates from Birth-Death Skyline traces over time.

    Parameters
    ----------
    traces_df : DataFrame
        Data frame of BEAST2 MCMC runs.
    partition_year_decimals : Series or list of dates
        Dates of partitioning skyline.
    youngest_tip_year_decimal : float
        Year decimal of youngest tip.
    parameter_start : str
        Starting string of parameters.
    x_tick_freq : str, default 'automatic'
        Frequency of x ticks. Options are 'automatic', 'yearly', 'quarterly', 'monthly', 'half month' or 'weekly'.
    style : str, default 'skyline'
        Style of plotting.
    y_label : str, default '$R_t$'
        Label of y-axis.
    palette : str, default sns.color_palette()
        Palette of plotting.
    include_grid : bool, default True
        Whether to include grid lines.
    grid_size : int, default 100
        Grid size for smoothing skyline interpolation.
    xml_set_label : str, default None
        Title for legend, if not provided 'xml_set' is used instead.

    Returns
    -------
    fig : Matplotlib figure
        Matplotlib figure.
    ax : Matplotlib axes
        Matplotlib axes.
    hdi_df : pd.DataFrame
        Table of HDIs with dates of boundaries.
    """
    if xml_set_label in traces_df.columns:
        xml_sets = traces_df[xml_set_label].unique()
    else:
        xml_sets = []
    if len(xml_sets) > 1:
        if partition_year_decimals is None:
            partition_year_decimals = {xml_set: None for xml_set in xml_sets}
        for name, value in {'partition_year_decimals': partition_year_decimals,
                            'youngest_tip_date': youngest_tip_year_decimal}.items():
            if not isinstance(value, dict):
                raise TypeError(name + ' must be a dict if there are multiple xml_sets.')
            if any(stain not in xml_sets for stain in value.keys()):
                raise ValueError('Each key of ' + name + ' should be the name of a xml_set.')
            if any(strain not in value.keys() for strain in xml_sets):
                raise ValueError('All of the xml_sets should be represented in ' + name + '.')
        plot_dfs = {}
        hdi_df = []
        for xml_set in xml_sets:
            if partition_year_decimals[xml_set] is not None:
                plot_df = _set_dates_skyline_plotting_df(traces_df[traces_df[xml_set_label] == xml_set],
                                                         parameter_start=parameter_start,
                                                         style=style,
                                                         partition_year_decimals=partition_year_decimals[xml_set])
                plot_dfs[xml_set] = deepcopy(plot_df)
                hdi_est = _plot_df_to_hdi_df(plot_df=plot_df)
                hdi_est.insert(0, xml_set_label, xml_set)
                hdi_df.append(hdi_est)
            else:
                if style != 'skyline':
                    raise ValueError('Currently only "skyline" style is supported when plotting without partition dates.')
                plot_df = _gridded_skyline(
                        traces_df[traces_df[xml_set_label] == xml_set],
                        parameter_start=parameter_start,
                        youngest_tip=youngest_tip_year_decimal[xml_set],
                        grid_size=grid_size)
                plot_dfs[xml_set] = deepcopy(plot_df)
                plot_df.insert(loc=0, column=xml_set_label, value=xml_set)
                hdi_df.append(plot_df)
        hdi_df = pd.concat(hdi_df)
    else:
        if partition_year_decimals is not None:
            plot_dfs = _set_dates_skyline_plotting_df(traces_df,
                                                      parameter_start=parameter_start,
                                                      style=style,
                                                      partition_year_decimals=partition_year_decimals)
            hdi_df = _plot_df_to_hdi_df(plot_dfs)
        else:
            if style != 'skyline':
                raise ValueError('Currently only "skyline" style is supported when plotting without partition dates.')
            plot_dfs = _gridded_skyline(
                traces_df,
                parameter_start=parameter_start,
                youngest_tip=youngest_tip_year_decimal,
                grid_size=grid_size)
            hdi_df = deepcopy(plot_dfs)

    if style.endswith('with mid-points'):
        x_vals = pd.concat([item[0] for item in plot_dfs.values()])['year_decimal']
    else:
        if isinstance(plot_dfs, dict):
            x_vals = pd.concat(plot_dfs.values())['year_decimal']
        else:
            x_vals = plot_dfs['year_decimal']
    tick_year_decimals, tick_labels = year_decimal_to_date_tick_labels(x_vals, tick_freq=x_tick_freq)
    fig, ax = plt.subplots(1, 1, layout='constrained', figsize=(7.5, 5))
    if not isinstance(plot_dfs, dict):
        plot_dfs = {None: plot_dfs}
    for i, (xml_set, plot_df) in enumerate(plot_dfs.items()):
        if style.endswith('with mid-points'):
            mid_point_df = plot_df[1]
            plot_df = plot_df[0]
            ax.scatter(mid_point_df['year_decimal'], mid_point_df['upper'], marker='v', c=palette[i])
            ax.scatter(mid_point_df['year_decimal'], mid_point_df['median'],marker='D', c=palette[i])
            ax.scatter(mid_point_df['year_decimal'], mid_point_df['lower'], marker="^", c=palette[i])

        ax.plot(plot_df['year_decimal'], plot_df['median'], color=palette[i], label=xml_set)
        ax.fill_between(plot_df['year_decimal'], plot_df['lower'], plot_df['upper'], color=palette[i],
                        alpha=0.2)

    ax.xaxis.set_ticks(tick_year_decimals)
    ax.set_xticklabels(tick_labels)
    ax.tick_params(axis='x', labelrotation=45)
    ax.set_xlabel('Date')
    ax.set_ylabel(y_label)
    if len(xml_sets) > 1:
        plt.legend()
    plt.tight_layout()
    if include_grid:
        plt.grid(linestyle='-.')

    return fig, ax, hdi_df

def outputs_for_possible_skyline(trace_df, youngest_tip_year_decimal, param_start, axis_label, grid_size=100, include_grid=True):
    """
    Determine if output figure should be a skyline if so plot with

    Parameters
    ----------
    trace_df : DataFrame
        Data frame of BEAST2 MCMC runs.
    youngest_tip_year_decimal : float
        Year decimal of youngest tip.
    param_start : str
        Starting string of parameters.
    axis_label:
    include_grid : bool, default True
        Whether to include grid lines.
    grid_size : int, default 100
        Grid size for smoothing skyline interpolation.

    Returns
    -------
    fig: matploplib figure
     ax: matploplib.axes
     hdi: pd.DataFrame
        Table of HD intervals
    """
    cols_starting_with = [col for col in trace_df.columns if col.startswith(param_start)]
    if len(cols_starting_with) > 1:
        fig, ax = plot_skyline(trace_df,
                               youngest_tip_year_decimal,
                               parameter_start=param_start,
                               y_label=axis_label,
                               grid_size=grid_size,
                               include_grid=include_grid)
        hdi = hdi_columns_starting_with(trace_df, param_start)
    elif len(cols_starting_with) == 1:
        fig, ax, hdi = plot_hist_kde(trace_df=trace_df, parameter=param_start, hdi_prob=0.95, x_label=axis_label)
    else:
        raise ValueError(f'Trace file does not contain any columns starting with "{param_start}".')
    return fig, ax, hdi


def plot_hist_kde(trace_df,
                  parameter,
                  x_label=None,
                  color=None,
                  hdi_prob=0.95,
                  tight_layout=True):
    """
    Plot histogram with kde.

    Parameters
    ----------
    trace_df: pandas.DataFrame
        DataFrame with column headed with parameters.
    parameter: str
        Parameter name.
    x_label: str, default None
        Label of x-axis. If None, use parameters name.
    color: str, default None
        Color of histogram and kde line.
    hdi_prob: float, default 0.95
        Prob for which the highest density interval will be computed. If not none
        lower and HDI lines plotted.
    tight_layout : bool, default True
        Whether to use tight layout or not.

    Returns
    -------
    if hdi_prob is None:
        Matplotlib figure and axis.
    else:
        Matplotlib figure, axis and dictionary of 'lower' and 'upper' hdi with median.
    """
    if x_label is None:
        x_label = parameter
    fig, ax = plt.subplots()  # initializes figure and plots
    sns.histplot(trace_df[parameter], ax=ax, kde=True, color=color)
    if hdi_prob is not None:
        hdi_est = hdi(trace_df[parameter].to_numpy(), hdi_prob=hdi_prob)
        upper_key = f'Upper {str(hdi_prob)} HDI'
        lower_key = f'Lower {str(hdi_prob)} HDI'
        hdi_est = {lower_key: hdi_est[0],
                   'Median': trace_df[parameter].median(),
                   upper_key: hdi_est[1]}
        ax.axvline(hdi_est['Median'], color='k', lw=2)
        ax.axvline(hdi_est[lower_key], color='k', ls='--', lw=1)
        ax.axvline(hdi_est[upper_key], color='k', ls='--', lw=1)
    ax.set_xlabel(x_label)
    if tight_layout:
        plt.tight_layout()

    if hdi_prob is None:
        return fig, ax
    else:
        return fig, ax, hdi_est


def plot_origin_or_tmrca(trace_df, parameter, x_tick_freq='automatic', hdi_prob=None, color=None):
    """Plot histograms of origin or TMRCA, with kde line.

    Parameters
    ----------
    trace_df: pandas.DataFrame
        DataFrame with column headed with parameters.
    parameter: str
        Parameter name.
    x_tick_freq : str, default 'automatic'
        Frequency of x ticks. Options are 'automatic', 'yearly', 'quarterly', 'monthly', 'half month' or 'weekly'.
    hdi_prob: float, default None
        Prob for which the highest density interval will be computed. If not none
        lower and HDI lines plotted.
    color: str or RGB tuple, default None
        Color of histogram and kde line.

    Returns
    -------
    if hdi_prob is None:
        Matplotlib figure and axis.
    else:
        Matplotlib figure, axis and dictionary of 'lower' and 'upper' hdi with median.
    """
    if parameter not in ['Origin', 'TMRCA']:
        raise ValueError('parameter must be "Origin" or "TMRCA".')
    tick_year_decimals, tick_labels = year_decimal_to_date_tick_labels(trace_df[parameter], tick_freq=x_tick_freq)
    output = plot_hist_kde(trace_df,
                           parameter,
                           x_label=f'{parameter} (date)',
                           color=color,
                           hdi_prob=hdi_prob,
                           tight_layout=False)
    if hdi_prob is None:
        fig, ax = output
    else:
        fig, ax, hdi_est = output
    ax.xaxis.set_ticks(tick_year_decimals)
    ax.set_xticklabels(tick_labels)
    ax.tick_params(axis='x', labelrotation=45)
    plt.tight_layout()
    if hdi_prob is None:
        return fig, ax
    else:
        hdi_est = {key: decimal_to_date(value) for key, value in hdi_est.items()}
        return fig, ax, hdi_est


def plot_comparative_origin_or_tmrca(df_melted, parameter, xml_set_label='xml_set',
                                     x_tick_freq='automatic', one_figure=False, palette=None):
    """Plot histograms of origins for each strain, with kde line.

    Parameters
    ----------
    df_melted : pandas.DataFrame
        Melted DataFrame for use in Seaborne.  Contains columns 'xml_set', 'variable' and
        'Estimate'.
    parameter : str
        Parameter to plot must be either 'Origin' or 'TMRCA'.
    xml_set_label: str, default None
        Legend title to use, if none provided this will be 'xml_set'.
    x_tick_freq: str, default 'automatic'
        Suggested x tick frequency. Options are 'automatic', 'yearly', 'quarterly', 
        'monthly', 'half month' or 'weekly'.
    one_figure: bool, default False
        If True, then one figure will be created with the histograms for each strain
        overlaid on top of each other.
        If False, then sub-figures will be created with a histogram for each strain per
        row.
    palette : list or dict of RGB colors (tuples), default None
        List of colors to use for each strain.
        Or dict of RGB colors to use for each strain.

    Returns
    -------
    If one_figure is True, then a single figure will be created.
    If one_figure is False, then a Seaborn.FactGrid will be created.
    """
    if parameter not in ['Origin', 'TMRCA']:
        raise ValueError('parameter must be "Origin" or "TMRCA".')
    df = df_melted[df_melted.variable == parameter]
    df.rename(columns={'Estimate': parameter}, inplace=True)
    tick_year_decimals, tick_labels = year_decimal_to_date_tick_labels(df[parameter], tick_freq=x_tick_freq)
    if one_figure:
        ax = sns.histplot(data=df, x=parameter, hue=xml_set_label, kde=True, palette=palette)
        if palette is None:
            palette = sns.color_palette()
        for colour, xml_set in zip(palette, df[xml_set_label].unique()):
            strain_df = df[df[xml_set_label] == xml_set]
            ax.axvline(strain_df[parameter].median(), color=colour)
            ax.axvline(strain_df[parameter].quantile(0.05), color=colour, ls='--', lw=1)
            ax.axvline(strain_df[parameter].quantile(0.95), color=colour, ls='--', lw=1)
        fig = ax
    else:
        fig = sns.FacetGrid(df, row=xml_set_label, hue=xml_set_label, margin_titles=True, aspect=4)
        fig.map_dataframe(sns.histplot, x=parameter, kde=True)
        fig.set_titles(row_template='')
        fig.add_legend()
        for ax, xml_set in zip(fig.axes.flat, df[xml_set_label].unique()):
            strain_df = df[df[xml_set_label] == xml_set]
            ax.axvline(strain_df[parameter].median(), color='k', lw=2)
            ax.axvline(strain_df[parameter].quantile(0.05), color='k', ls='--', lw=1)
            ax.axvline(strain_df[parameter].quantile(0.95), color='k', ls='--', lw=1)

    ax.xaxis.set_ticks(tick_year_decimals)
    ax.set_xticklabels(tick_labels)
    ax.tick_params(axis='x', labelrotation=45)
    plt.tight_layout()
    return fig



"""
This code has for functions timescale and write_nodes_CI been adapted from
 https://gitlab.in2p3.fr/ete/CoV-flow/-/blob/master/scripts/tree_time_scale.py?ref_type=heads.
On 2024-01-13

"""
import copy
from treetime import TreeTime
import numpy as np
from treetime.utils import parse_dates
import pandas as pd
import matplotlib.pyplot as plt
from beast_pype.fig_utils import year_decimal_to_date_tick_labels



def timescale(ftree, falignment, fdates, reroot='least-squares', clock_rate=None, clock_std=None,
              clock_filter=None, remove_root=True, coalescent_tc="opt",
              sample_id_field = None,
              collection_date_field = 'date',
              rng_seed=None,
               **kwargs):
    """
    Timescale a phylogenetic tree using tree time.

    Parameters
    ----------
    ftree: str
        Path to newick tree file.
    falignment: str
        Path to fasta_path alignment file.
    fdates: str
        Path to dates file.
    reroot: str or list of str, default `least-squares` 
        Method to reroot the tree see `treetime.TreeTime.reroot`:
        Which method should be used to find the best root. Available methods are:

            :code:`best`, `least-squares` - minimize squared residual or likelihood of
             root-to-tip regression.

            :code:`min_dev` - minimize variation of root-to-tip distance.

            :code:`oldest` - reroot on the oldest node.

            :code:`<node_name>` - reroot to the node with name :code:`<node_name>`.

            :code:`[<node_name1>, <node_name2>, ...]` - reroot to the MRCA of these
             nodes.
    clock_rate: float
        Mutation rate (substitutions per position per year).
    clock_std: float
        Standard deviation of the mutation (rates substitutions per position per year).
    clock_filter: float or None
        If given clock filter applies (treetime.TreeTime.clock_filter).
        This value is then used in
         n_iqd:  float
            The number of iqd intervals. The outlier nodes are those which do not fall
             into :math:`IQD\cdot n_iqd` interval (:math:`IQD` is the interval between
            75\ :sup:`th` and 25\ :sup:`th` percentiles)
    remove_root: bool, default True
    coalescent_tc: : float, str
           Value used in
           If not None, use coalescent model to correct the branch lengths by
           introducing merger costs.
           If Tc is float, it is interpreted as the coalescence timescale.
           If Tc is str, it should be one of (:code:`opt`, :code:`const`, :code:`skyline`)
    sample_id_field : str, optional
        Name of column containing taxon names in fdates. If None, will use
        first column that contains 'name', 'strain', 'accession'
    collection_date_field : str, default='date'
        Name of column containing taxon names in fdates. If None, will use
        a column that contains the substring 'date'.
    rng_seed: int, optional
        Seed for random number generator.

    kwargs: dict, default None
        Key word arguments to pass to TreeTime.run.


    Returns
    -------
    time_tree: treetime.TreeTime
    bad_tips : list of str
    """
    bad_tips = list()
    dates = parse_dates(fdates, name_col=sample_id_field, date_col=collection_date_field)

    time_tree = TreeTime(gtr='JC69', tree=ftree, aln=falignment, dates=dates,
                         verbose=1, use_fft=True, precision='auto', rng_seed=rng_seed)

    if clock_filter is not None:
        time_tree.clock_filter(reroot=reroot, n_iqd=clock_filter, plot=False)
        tips = [x for x in time_tree.tree.get_terminals()]
        for n in tips:
            if n.bad_branch:
                time_tree.tree.prune(n)
                bad_tips.append(n.name)
        time_tree.prepare_tree()
        # remove bad tips
        if len(bad_tips):
            print("Pruning leaves :\n", "\n".join(bad_tips))
            with open('results/treetime_ignored_tips.txt', 'w') as f:
                for tip in bad_tips:
                    count = f.write("%s\n" % tip)

        time_tree.stl = {n.name: n for n in time_tree.tree.get_terminals()}
        time_tree.stl.update({n.name.upper(): n for n in time_tree.tree.get_terminals()})
        time_tree.stl.update({n.name.lower(): n for n in time_tree.tree.get_terminals()})
        time_tree.tree.root.up = None
        for node in time_tree.tree.get_nonterminals():
            for n in node.clades:
                n.up = node
    marginal = 'always'  # 'assign' # estimate confidence intervals via marginal ML and assign
    branch_length_inference = 'joint'  # auto
    resolve_polytomies = True
    max_iter = 2
    covariance = False

    time_tree.run(infer_gtr=True, root=reroot, Tc=coalescent_tc, time_marginal=marginal,
                  branch_length_mode=branch_length_inference, resolve_polytomies=resolve_polytomies,
                  max_iter=max_iter, fixed_pi=None, fixed_clock_rate=clock_rate,
                  stochastic_resolve=resolve_polytomies, vary_rate=clock_std, use_covariation=covariance,
                  raise_uncaught_exceptions=True, **kwargs)

    if remove_root and (reroot != 'least-squares' or reroot != 'best'):
        for root in reroot:
            if root not in bad_tips:
                root = root.strip()
                time_tree.tree.prune(root)

    time_tree.convert_dates()
    time_tree.branch_length_to_years()

    return time_tree, bad_tips


def tree_nodes_ci(time_tree, fraction=0.95):
    """
    Get node confidence intervals from tree time tree.

    Parameters
    ----------
    time_tree: treetime.TreeTime
    fraction: float
        Confidence interval fraction

    Returns
    -------
    pd.Dataframe
    """
    records = []
    for n in time_tree.tree.find_clades():
        conf = time_tree.get_max_posterior_region(n, fraction=fraction)
        record = {
            'node': n.name,
            'date': n. date,
            'year_decimal': n.numdate,
            'interval_' + str(1-fraction): conf[0],
            'upper_' + str(fraction): conf[1]
        }
        records.append(record)
    return pd.DataFrame.from_records(records)

def plot_root_to_tip(time_tree, label=True, x_tick_freq='automatic'):
    """
    Plot root-to-tip regression.

    Parameters
    ----------
    time_tree: time_tree: treetime.TreeTime
    label: bool, default True
        If true, label the plot.
    x_tick_freq: str, default='automatic'
        Suggested tick frequency. Options are 'automatic', 'yearly', 'quarterly',
        'monthly', 'half month' or 'weekly'.

    Returns
    -------
    fig, ax: matplotlib.figure.Figure
    """
    fig, ax = plt.subplots(1, 1)
    time_tree.plot_root_to_tip(ax=ax, label=label)
    x_year_decimal = np.array([tip.raw_date_constraint
                               for tip in time_tree.tree.get_terminals()])
    tick_year_decimals, tick_labels = year_decimal_to_date_tick_labels(x_year_decimal,
                                                                       tick_freq=x_tick_freq)
    ax.xaxis.set_ticks(tick_year_decimals)
    ax.set_xticklabels(tick_labels)
    ax.tick_params(axis='x', labelrotation=45)
    return fig, ax



def temporal_pruning_sampler(time_tree: TreeTime, sample_size: int, draws=1, seed=None):
    """
    Sample a time tree object.

    Removal of tips is based on normalised residual of temporal signal.
    Parameters
    ----------
    time_tree : treetime.TreeTime
        Time tree to remove tips from.
    sample_size: int
        Sample size of desired tree.
    draws : int
        Number of draws.
    seed : int
        Random seed for sampling.

    Returns
    -------
   If draws>1:
        A list of lists of strain ids/names.
   If draws==1:
        A list of strain ids/names.
    """
    if not isinstance(draws, int):
        raise ValueError("draws must be an integer > 0.")
    tips = time_tree.tree.get_terminals()
    tip_names = [tip.name for tip in tips]
    n_tips = len(tips)
    to_prune = n_tips - sample_size
    if to_prune == 0:
        raise ValueError('time_tree provided has tips equal to sample_size')
    if to_prune < 0:
        raise ValueError('time_tree provided has less tips than sample_size')
    # get values of terminals
    x_year_decimal = np.array([tip.raw_date_constraint for tip in tips])
    root_to_tip_actual = np.array([tip.dist2root for tip in tips])
    root_to_tip_expected = time_tree.clock_model['slope'] * x_year_decimal + time_tree.clock_model['intercept']
    abs_residuals = np.absolute(root_to_tip_actual - root_to_tip_expected)
    prune_prob = abs_residuals / abs_residuals.sum()
    selections = []
    rng = np.random.default_rng(seed=seed)
    for draw in range(draws):
        selection = copy.deepcopy(tip_names)
        to_prune_index = rng.choice(n_tips, size=to_prune, p=prune_prob, replace=False)
        selection = [ele for index, ele in enumerate(selection) if index not in to_prune_index]
        selections.append(selection)

    if draws==1:
        return selections[0]
    else:
        return selections








from base_test_classes import SimpleWorkflowVariationTest

xml_generation_notebook = 'Phase-3-Gen-Generic-xml.ipynb'
workflow = 'Generic'

class TestGenericFull(SimpleWorkflowVariationTest):
    name = 'Test_Generic_Full'
    parameters_path = 'tests/parameters/Generic_full.yml'
    workflow = workflow
    initial_tree_building = True
    xml_generation = True
    downsampling = True
    xml_generation_notebook = xml_generation_notebook

class TestGenericNoDownsampling(SimpleWorkflowVariationTest):
    name = 'Test_Generic_No_Downsampling'
    parameters_path = 'tests/parameters/Generic_no_downsampling.yml'
    workflow = workflow
    initial_tree_building = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestGenericNoInitialTree(SimpleWorkflowVariationTest):
    name = 'Test_Generic_No_Initial_Tree'
    parameters_path = 'tests/parameters/Generic_no-initial-tree.yml'
    workflow = workflow
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestGenericXmlReadyToGo(SimpleWorkflowVariationTest):
    name = 'Test_Generic_XML_Ready_To_Go'
    parameters_path = 'tests/parameters/Generic_xml-ready-to-go.yml'
    workflow = workflow
    xml_generation_notebook = xml_generation_notebook


from beast_pype.date_utilities import decimal_to_date, date_to_decimal
import pandas as pd


def year_decimal_to_date_tick_labels(series, tick_freq='automatic'):
    """
    Uses year_decimal values to propose date tick labels.

    Parameters
    ----------
    series: pd.Series
        Series of year_decimal values.
    tick_freq: str, default='automatic'
        Suggested tick frequency. Options are 'automatic', 'yearly', 'quarterly',
        'monthly', 'half month' or 'weekly'.

    Returns
    -------
    tick_year_decimals: pd.Series
        Location of ticks. A Series of tick year_decimal values.
    tick_labels:
        The corresponding tick labels.
    """
    max_date = decimal_to_date(series.max())
    min_date = decimal_to_date(series.min())
    if tick_freq == 'automatic':
        time_length = series.max() - series.min()
        if time_length > 5:
            tick_freq = 'yearly'
        elif time_length > 2:
            tick_freq = 'quarterly'
        elif time_length > 1:
            tick_freq = 'monthly'
        elif time_length > 0.5:
            tick_freq = 'half month'
        else:
            tick_freq = 'weekly'

    if tick_freq == 'yearly':
        freq = 'YS'
        strftime = '%Y'
    elif tick_freq == 'quarterly':
        freq = 'QS'
        strftime = '%Y-%b'
    elif tick_freq == 'monthly':
        freq = 'MS'
        strftime = '%Y-%b'
    elif tick_freq == 'half month':
        freq = 'SME'
        strftime = '%Y-%b-%d'
    elif tick_freq == 'weekly':
        freq = 'W'
        strftime = '%Y-%b-%d'
    else:
        raise ValueError \
            ('x_tick_freq must be "automatic", "yearly", "quarterly", "monthly", ''half month" or "weekly".' +
                         '\n x_tick_freq=' + str(tick_freq))

    tick_dates = pd.date_range(start=min_date.replace(day=1), end=max_date, freq=freq)
    tick_labels = tick_dates.to_series().dt.strftime(strftime).to_list()
    tick_year_decimals = tick_dates.map(date_to_decimal)
    return tick_year_decimals, tick_labels


"""
Functions for the generation of beast xmls for specific BEAST models
"""
from dark.fasta import FastaReads
from beast2xml.beast2 import BEAST2XML
import ete3
import pandas as pd
from beast_pype.date_utilities import date_to_decimal
import warnings

def two_df_cols_to_dict(df, key, value):
    return df[[key, value]].set_index(key).to_dict()[value]

def gen_xml_from_any_template(template_xml_path,
                              fasta_path,
                              metadata_path,
                              output_path,
                              sample_id_field='strain',
                              collection_date_field='date',
                              initial_tree_path=None,
                              log_file_basename=None,
                              chain_length=None,
                              trace_log_every=None,
                              tree_log_every=None,
                              screen_log_every=None,
                              store_state_every=None):
    """
    Generate a BEAST 2 xml from a BEAST 2 xml template for any model.

    Parameters
    ----------
    template_xml_path: str
        Path to template_xml_path.
    fasta_path: str
        Path to sequences must be a fasta file.
    metadata_path: str
        Path to metadata_update must be csv.
    sample_id_field: str, default 'strain'
        Field used to identify samples.
    collection_date_field: str, default 'date'
        Field to use as sequence collection date.
    output_path: str
        Path to save output xml to.
    initial_tree_path: str, optional
        Path to initial_tree    must be Newick file (nwk).
    log_file_basename: str, optional
            The base filename to write logs to. A .log or .trees suffix will be appended
            to this to make the actual log file names.  If None, the log file names in
            the template will be retained.
    chain_length : int, optional
        The length of the MCMC chain. If C{None}, the value in the template will
         be retained.
    trace_log_every: int, optional
        Specifying how often to write to the trace log file. If None, the value in the
        template will be retained.
    tree_log_every: int, optional
        Specifying how often to write to the file_path log file. If None, the value in the
        template will be retained.
    screen_log_every: int, optional
        Specifying how often to write to the terminal (screen) log. If None, the
        value in the template will be retained.
    store_state_every : int, optional
        Specifying how often to write MCMC state file. If None, the
        value in the template will be retained.

    """
    if metadata_path.endswith('.tsv'):
        delimiter = '\t'
    elif metadata_path.endswith('.csv'):
        delimiter = ','
    else:
        raise TypeError(
            f"metadata_path must be a csv or tsv file, ending with the appropriate file extension. Value given is {metadata_path}")
    beast2xml = BEAST2XML(template=template_xml_path)
    seqs = FastaReads([fasta_path])
    beast2xml.add_sequences(seqs)
    beast2xml.add_dates(date_data=metadata_path,
                        seperator=delimiter,
                        sample_id_field=sample_id_field,
                        collection_date_field=collection_date_field)
    if initial_tree_path is not None:
        beast2xml.add_initial_tree(initial_tree_path)
    beast2xml.to_xml(
        output_path,
        chain_length=chain_length,
        log_file_basename=log_file_basename,
        trace_log_every=trace_log_every,
        tree_log_every=tree_log_every,
        screen_log_every=screen_log_every,
        store_state_every=store_state_every,
    )


def gen_bdsky_serial_xml(template_xml_path,
                         fasta_path,
                         metadata_path,
                         output_path,
                         collection_date_field='date',
                         sample_id_field='strain',
                         initial_tree_path=None,
                         origin_upper_height_addition=None,
                         origin_start_addition=None,
                         origin_prior=None,
                         rt_dims=None,
                         rt_change_dates=None,
                         sampling_prop_dims=None,
                         sampling_prop_change_dates=None,
                         zero_sampling_before_first_sample=True,
                         log_file_basename=None,
                         chain_length=None,
                         trace_log_every=None,
                         tree_log_every=None,
                         screen_log_every=None,
                         store_state_every=None):
    """
    Generate a BDSKY xml with initial tree inserted.

    Parameters
    ----------
    template_xml_path: str
        Path to template_xml_path.
    fasta_path: str
        Path to sequences must be a fasta file.
    metadata_path: str
        Path to metadata_update must be csv.
    output_path: str
        Path to save output xml to.
    sample_id_field: str, default 'strain'
        Field used to identify samples.
    collection_date_field: str, default 'date'
        Field to use as sequence collection date.
    initial_tree_path: str, optional
        Path to initial_tree    must be Newick file (nwk).
    origin_upper_height_addition: int or float, optional
        Value to add to tree height for upper limit of origin prior. Origin prior is
         uniformly distributed.
    origin_start_addition: int or float, optional
        Value to add to tree height for starting value. . Origin prior is uniformly
         distributed.
    origin_prior: dict {'lower': float, 'upper': float, 'start': float}, optional
        Details of the origin prior assumed to be uniformly distributed.
    rt_dims: int, optional
        Number of Rt dimensions (time periods).
    rt_change_dates: : list, tuple, pd.Series or pd.DatetimeIndex of datetimes
        Internal partitions of Rt estimation periods.
    sampling_prop_dims: int, optional
        Number of sampling proportion dimensions (time periods).
    sampling_prop_change_dates: : list, tuple, pd.Series or pd.DatetimeIndex of dates
        Internal partitions of sampling proportion estimation periods.
    zero_sampling_before_first_sample: bool, default True
        Whether to have zero sampling before date the first sample was collected.
    log_file_basename: str, optional
            The base filename to write logs to. A .log or .trees suffix will be appended
            to this to make the actual log file names.  If None, the log file names in
            the template will be retained.
    chain_length : int, optional
        The length of the MCMC chain. If C{None}, the value in the template will
         be retained.
    trace_log_every: int, optional
        Specifying how often to write to the trace log file. If None, the value in the
        template will be retained.
    tree_log_every: int, optional
        Specifying how often to write to the file_path log file. If None, the value in the
        template will be retained.
    screen_log_every: int, optional
        Specifying how often to write to the terminal (screen) log. If None, the
        value in the template will be retained.
    store_state_every : int, optional
        Specifying how often to write MCMC state file. If None, the
        value in the template will be retained.

    """
    if metadata_path.endswith('.tsv'):
        delimiter = '\t'
    elif metadata_path.endswith('.csv'):
        delimiter = ','
    else:
        raise TypeError(
            f"metadata_path must be a csv or tsv file, ending with the appropriate file extension. Value given is {metadata_path}")
    metadata_df = pd.read_csv(metadata_path, parse_dates=[collection_date_field], sep=delimiter)
    metadata_df['year_decimal'] = metadata_df[collection_date_field].map(date_to_decimal)
    if origin_prior is None:
        if origin_upper_height_addition is not None and  origin_start_addition is not None:
            if initial_tree_path is None:
                raise ValueError('If parameterising the origin prior via ' +
                                 'origin_upper_height_addition and origin_start_addition an ' +
                                 'initial tree must be provided.')
            tree = ete3.Tree(initial_tree_path, format=1)
            furthest_leaf, tree_height = tree.get_farthest_leaf()
            youngest_tip = metadata_df.year_decimal.max()
            oldest_tip = metadata_df.year_decimal.min()
            tip_distance = youngest_tip - oldest_tip
            if tip_distance > tree_height:
                raise ValueError('tree_height must be greater than distance between youngest_tip_date and oldest_tip.')
            origin_prior = {
                'lower': tip_distance,
                'upper': tree_height + origin_upper_height_addition,
                'start': tree_height + origin_start_addition}
    else:
        warnings.warn("If using your own origin prior there is a chance" +
                      " that an origin value will be less than the tree " +
                      " height when BEAST 2 is running." +
                      " If this happens BEAST 2 will crash.\n"+
                      "We recommend using one generated by supplying:\n" +
                      "* An initial temporal tree. \n" +
                      "* origin_upper_height_addition \n" +
                      "* origin_start_addition")
    beast2xml = BEAST2XML(template=template_xml_path)
    seqs = FastaReads([fasta_path])
    beast2xml.add_sequences(seqs)
    beast2xml.add_dates(date_data=metadata_path,
                        seperator=delimiter,
                        sample_id_field=sample_id_field,
                        collection_date_field=collection_date_field)
    if origin_prior is not None:
        # Change Origin starting value, lower and upper limit on state node
        beast2xml.change_parameter_state_node("origin", value=origin_prior["start"])
        del origin_prior["start"]
        beast2xml.change_prior("origin", "uniform", **origin_prior)
    else:
        warnings.warn("If using the Origin prior in the template xml there is a chance" +
                      " that an origin value will be less than the tree " +
                      " height when BEAST 2 is running.." +
                      " If this happens BEAST 2 will crash.\n"+
                      "We recommend using one generated by supplying:\n" +
                      "* An initial temporal tree. \n" +
                      "* origin_upper_height_addition \n" +
                      "* origin_start_addition")
    if rt_change_dates is not None:
        if rt_dims is not None:
            raise AssertionError("Either rt_dims or rt_partitions can be given but not both.")
        beast2xml.add_rate_change_dates(
            parameter="birthRateChangeTimes",
            dates=rt_change_dates)
    if rt_dims is not None:
        beast2xml.change_parameter_state_node(parameter='reproductiveNumber',
                                              dimension=rt_dims)
    if sampling_prop_change_dates is not None:
        if sampling_prop_dims is not None:
            raise AssertionError("Either sampling_prop_dims or sampling_prop_partition_freq can be given but not both.")
        if zero_sampling_before_first_sample:
            beast2xml.add_rate_change_dates(
                parameter="samplingRateChangeTimes",
                dates=sampling_prop_change_dates,
                offset_earliest=1e-6)
            beast2xml.set_dimension_values_to_0(parameter="samplingProportion")
        else:
            beast2xml.add_rate_change_dates(
                parameter="samplingRateChangeTimes",
                dates=sampling_prop_change_dates)
    if sampling_prop_dims is not None:
        beast2xml.change_parameter_state_node(parameter='samplingProportion',
                                              dimension=sampling_prop_dims)
    if initial_tree_path is not None:
        beast2xml.add_initial_tree(initial_tree_path)
    beast2xml.to_xml(
        output_path,
        chain_length=chain_length,
        log_file_basename=log_file_basename,
        trace_log_every=trace_log_every,
        tree_log_every=tree_log_every,
        screen_log_every=screen_log_every,
        store_state_every=store_state_every,
    )


from datetime import timedelta, datetime
from datetime import datetime as dt
import calendar
from dateutil.relativedelta import relativedelta, SU, MO
import time


def _since_epoch(date):  #
    """
    Returns seconds since epoch

    Parameters
    ----------
    date: datetime.date or datetime.datetime
        Date to be converted.

    Returns
    -------
    float
    """
    return time.mktime(date.timetuple())


def decimal_to_date(decimal):
    """
    Converts a decimal year to a date object.

    Parameters
    ----------
    decimal: float
        The decimal value to convert.

    Returns
    -------
    date: datetime.date
    """
    year = int(decimal)
    d = timedelta(days=(decimal - year) * (365 + calendar.isleap(year)))
    day_one = datetime(year, 1, 1)
    date = d + day_one
    return date


def date_to_decimal(date):
    """ Convert date to year decimal (year fraction).

    Parameters
    ----------
    date: datetime.date or datetime.datetime
        Date to be converted.

    Returns
    -------
    date_as_year_decimal : float
    """
    year = date.year
    start_of_this_year = dt(year=year, month=1, day=1)
    start_of_next_year = dt(year=year + 1, month=1, day=1)

    year_elapsed = _since_epoch(date) - _since_epoch(start_of_this_year)
    year_duration = _since_epoch(start_of_next_year) - _since_epoch(start_of_this_year)
    fraction = year_elapsed / year_duration

    return date.year + fraction


def the_next_sunday(date, dont_if_sunday=False):
    """
    Find the next Sunday after a given date.

    Parameters
    ----------
    date: datetime.date or datetime.datetime
    dont_if_sunday: bool, default False
        If already a Sunday, return date.

    Returns
    -------
    datetime.datetime: The next Sunday after date.
    """
    if dont_if_sunday and date.weekday()==6:
        return date
    else:
        return date + relativedelta(weekday=SU(1))

def the_previous_monday(date, dont_if_monday=False):
    """
    Find the previous Monday before a given date.

    Parameters
    ----------
    date: datetime.date or datetime.datetime
    dont_if_monday: bool, default False
        If already a Monday, return date.

    Returns
    -------
    datetime.datetime: The previous Monday before date.
    """
    if dont_if_monday and date.weekday()==0:
        return date
    else:
        return date + relativedelta(weekday=MO(-1))


import nbformat as nbf
import os
from beast_pype.path_utils import path_to_workflow_modules, path_to_report_templates
from papermill.iorw import load_notebook_node, write_ipynb
from papermill.parameterize import parameterize_notebook
from papermill.inspection import _infer_parameters
from papermill.iorw import load_notebook_node



workflow_modules_path = path_to_workflow_modules()
reports_path =  path_to_report_templates()
available_reports = [file.name.replace('.ipynb', '') for file in reports_path.glob('*.ipynb')]

def gen_xml_set_diag_notebook(save_dir,
                              directories_to_exclude=None):
    """
    Generates rest of Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb.

    Parameters
    ----------
    save_dir: str
        Path to save copy of Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb.
        Each subdirectory of save_dir will be assumed to contain data of a xml_set
         and used in generating the new Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb.
    directories_to_exclude:
        Any directories to exclude from save_dir when generating new
        Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb.

    Returns
    -------
    None
    """
    if directories_to_exclude is None:
        directories_to_exclude = ['.ipynb_checkpoints']
    directories = [directory for directory in os.listdir(save_dir)
                   if os.path.isdir(os.path.join(save_dir, directory)) and
                   directory not in directories_to_exclude]
    diag_notebook_path = f'{workflow_modules_path}/Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb'
    diag_notebook = nbf.read(diag_notebook_path, as_version=4)
    for xml_set in directories:
        diag_notebook['cells'] += [
            nbf.v4.new_markdown_cell(
                f"## Diagnosing XML Set: {xml_set}\n" +
                "### Loading data"),
            nbf.v4.new_code_cell(
                f"if os.path.exists('{xml_set}/beast_outputs'):\n" +
                f"\tbeast_outputs_path = '{xml_set}/beast_outputs'\n" +
                "else:\n" +
                f"\tbeast_outputs_path = '{xml_set}'\n\n"
                "sample_diag = BEASTDiag(beast_outputs_path)\n"+
                f"beast_outputs_paths['{xml_set}'] = beast_outputs_path"
            ),
            nbf.v4.new_markdown_cell(
                "## Selecting burn-in and Chains to Remove\n\n" +
                "Activating the cell below will generate an interactive widget. Widgets parts:\n" +
                "* Top interactive part: this allows you to select for a different burn-in and remove chains and select the parameters used in the rest of the widget.,\n" +
                "* Middle display: KDE and trace plots, see [arviz.plot_trace documentation](https://python.arviz.org/en/stable/api/generated/arviz.plot_trace.html#arviz.plot_trace).\n" +
                "* Bottom display: A table of statistics regarding the traces, see [arviz.summary documentation](https://python.arviz.org/en/stable/api/generated/arviz.summary.html#arviz.summary). Regarding these statistics:\n" +
                "\t* Ideally the ESSs should be >= 200, see [arviz.ess documentation](https://python.arviz.org/en/stable/api/generated/arviz.ess.html#arviz.ess).\n" +
                "\t* Ideally the r_hat should be close fo 1 (<1.05), see [arviz.rhat documentation](https://python.arviz.org/en/stable/api/generated/arviz.rhat.html#arviz.rhat).\n" +
                "\t* Markov Chain Standard Error MCSEs, see [arviz.mcse](https://python.arviz.org/en/stable/api/generated/arviz.mcse.html#arviz.mcse).\n\n"+
                "After making your selection click on the cell below the widget and then keep pressing shift+enter to carry on with the rest of the cells in this notebook."
            ),
            nbf.v4.new_code_cell(
                "sample_diag_widget = sample_diag.generate_widget(parameters_displayed=4)\n" +
                "sample_diag_widget"),
            nbf.v4.new_code_cell(
                f"pipeline_run_info['Chains Used']['{xml_set}'] = deepcopy(sample_diag.selected_chains)\n" +
                f"pipeline_run_info['Burn-In']['{xml_set}'] = deepcopy(sample_diag.burinin_percentage)\n" +
                f"phase_5i_params = sample_diag.merging_outputs_params(output_path=outputs_and_reports_dir, kernel_name=kernel_name, xml_set='{xml_set}')\n" +
                "phase_5i_log = execute_notebook(input_path=f'{workflow_modules_path}/Phase-5i-Merge-BEAST-outputs.ipynb',\n" +
                 f"output_path=save_dir + '/{xml_set}/Phase-5i-Merge-BEAST-outputs.ipynb',\n" +
                "\t\t\t\t\t\t\t\tparameters=phase_5i_params,\n" +
                "\t\t\t\t\t\t\t\tprogress_bar=True,\n" +
                "\t\t\t\t\t\t\t\tnest_asyncio=True, start_timeout=600)"
            )
        ]

    diag_notebook['cells'] += [
        nbf.v4.new_markdown_cell(
            "## Update the pipeline_run_info yaml."),
        nbf.v4.new_code_cell(
            "with open(f'{save_dir}/pipeline_run_info.yml', 'w') as fp:\n\tyaml.dump(pipeline_run_info, fp, sort_keys=True)\nfp.close()"
        ),
        nbf.v4.new_markdown_cell('## Get BEAST 2 runtimes'),
        nbf.v4.new_code_cell(
            "runtimes_dfs = []\n"+
            "job_stats_dfs = []\n"
            "for xml_set, path in beast_outputs_paths.items():\n" +
            "\tentry_df = get_beast_runtimes(path, outfile_startswith='run-with-seed-', outfile_endswith='.out')\n"
            "\tentry_df['xml_set'] = xml_set\n" +
            "\truntimes_dfs.append(entry_df)\n" +
            "\tif os.path.isfile(f'{path}/slurm_job_ids.txt'):\n" +
            "\t\tjobs_df = pd.read_csv(f'{path}/slurm_job_ids.txt', sep=';')\n" +
            "\t\tjobs_df['JobID']=jobs_df['JobID'].astype(str)\n" +
            "\t\tstats_df = get_slurm_job_stats(jobs_df['JobID'].to_list())\n" +
            "\t\tjob_stats_entry = jobs_df.merge(stats_df, on='JobID')\n" +
            "\t\tjob_stats_entry['xml_set'] = xml_set\n" +
            "\t\tjob_stats_dfs.append(job_stats_entry)\n" +
            "runtimes_df = pd.concat(runtimes_dfs)\n"+
            "runtimes_df.to_csv(f'{outputs_and_reports_dir}/BEAST_runtimes.csv', index=False)\n"+
            "if job_stats_dfs:\n"+
            "\tjob_stats_df = pd.concat(job_stats_dfs)\n" +
            "\tjob_stats_df.to_csv(f'{outputs_and_reports_dir}/BEAST_slurm_stats.csv', index=False)"
        ),
        nbf.v4.new_markdown_cell(
            '## Generate output Report\n' +
            'Now you can now move on to visualising outputs from BEAST using a report template.'
        ),
        nbf.v4.new_code_cell(
            "report_params = {'save_dir': outputs_and_reports_dir, 'beast_xml_path':beast_xml_path, 'xml_set_label': xml_set_label}\n" +
            "output_report_path = f'{outputs_and_reports_dir}/BEAST_pype-Report.ipynb'\n" +
            "add_unreported_outputs(report_template, outputs_and_reports_dir, output_report_path, xml_set_comparisons=True)\n" +
            "output = execute_notebook(\n" +
            "\tinput_path=output_report_path,\n" +
            "\toutput_path=output_report_path,\n" +
            "\tparameters=report_params,\n" +
            "\tkernel_name=kernel_name,\n" +
            "\tprogress_bar=True)"),
        nbf.v4.new_markdown_cell(
            "### Convert Output Report from Jupyter Notebook to Notebook\n" +
            "This also removes code cells."
        ),
        nbf.v4.new_code_cell(
            "%%bash -l -s {output_report_path}\n" +
            "source activate beast_pype\n" +
            "jupyter nbconvert --to html --no-input $@"
        ),
        nbf.v4.new_markdown_cell(
            "## Produce MCC trees\n\n"+
            "This can take sometime. So you can look at the report whilst waiting this is done last."
        ),
        nbf.v4.new_code_cell(
            "gen_mcc_notebook(outputs_and_reports_dir, 'Phase-5ii-Gen-MCC-Trees.ipynb', kernel_name=kernel_name)\n" +
            "mcc_tree_output = execute_notebook(input_path='Phase-5ii-Gen-MCC-Trees.ipynb',\n" +
            "output_path='Phase-5ii-Gen-MCC-Trees.ipynb',\n" +
            "progress_bar=True)\n"
        )
    ]

    with open(f'{save_dir}/Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb', 'w') as f:
        nbf.write(diag_notebook, f)

def gen_beast_diagnostic_nb(beast_outputs,
                            report_template,
                            kernel_name='beast_pype',
                            beast_xml_path=None, **kwargs):
    """
    Generate notebook for Diagnosing BEAST 2 MCMC runs & generating a report.

    Parameters
    ----------
    beast_outputs: str
        A valid path containing output `.log` and `.tree` files from running BEAST 2. The
        Phase 5 diagnostic report is generated within this directory.
        On how BEAST 2 outputs should be organised:
        *  If report_template is for simple reports ('BDSKY-Serial',
            or 'Generic') this should contain BEAST 2 outputs you wish to diagnose.
        *   If report_template is for comparative reports ('BDSKY-Serial-Comparative',
            'COVID-Strain-Surveillance' or 'Generic-Comparative') this should contain
            subdirectories each containing the BEAST 2 outputs you wish to diagnose and compare.
        Note in both cases beast outputs can be contained in a subdirectory labeled `beast_outputs`.
    report_template: str
        Name of a valid report template to use to generate report.
    kernel_name: str, default 'beast_pype'
        Name of Jupyter python kernel_name to use when running diagnostic & report template notebooks.
    beast_xml_path: str or dict of strings, optional
        Path(s) to BEAST 2 xml used to generate BEAST outputs. If not given, the
        location described in beast_outputs is checked for the file `beast.xml`
        and that is used instead.
        If report_template is for comparative reports ('BDSKY-Serial-Comparative',
        'COVID-Strain-Surveillance' or 'Generic-Comparative') this should be a dictionary
        with keys being the names of xml_sets (names of subdirectories described in
        beast_outputs) values being the paths to the corresponding BEAST 2 xml file.



    """
    if report_template not in available_reports:
        raise ValueError('The value for report_template must be one of the following: \n{}'.format(', '.join(available_reports)))
    report_template_path = f"{reports_path}/{report_template}.ipynb"
    report_nb = load_notebook_node(report_template_path)
    accepted_parameters = set([parameter.name
                           for parameter in _infer_parameters(report_nb)])
    invalid_params = set(kwargs.keys()) - accepted_parameters
    if invalid_params:
        raise ValueError(
            f'The following parameters that have been supplied for {report_template} as kwargs are invalid: \n {", ".join(invalid_params)}')
    parameters = kwargs
    comparative_reports = ['BDSKY-Serial-Comparative',
                           'COVID-Strain-Surveillance',
                           'Generic-Comparative']
    if report_template in comparative_reports:
        if beast_xml_path is None:
            beast_xml_path = {}
            for entry in os.listdir(beast_outputs):
                if os.path.isdir(f"{beast_outputs}/{entry}") and entry != '.ipynb_checkpoints':
                    beast_xml_path_value = f"{entry}/beast.xml"
                    if not os.path.isfile(f"{beast_outputs}/{beast_xml_path_value}"):
                        beast_xml_path_value = f"{entry}/beast_outputs/beast.xml"
                    beast_xml_path[entry] = beast_xml_path_value
        gen_xml_set_diag_notebook(beast_outputs, directories_to_exclude=['.ipynb_checkpoints'])
        diagnostic_save_name = f"{beast_outputs}/Phase-5-Diagnosing-XML-sets-and-Generate-Report.ipynb"
        diagnostic_nb = load_notebook_node(diagnostic_save_name)
    else:
        if beast_xml_path is None:
            beast_xml_path = "beast.xml"
            if not os.path.isfile(f"{beast_outputs}/{beast_xml_path}"):
                beast_xml_path = "beast_outputs/beast.xml"
        diagnostic_save_name = f'{beast_outputs}/Phase-5-Diagnosing-Outputs-and-Generate-Report.ipynb'
        diagnostic_nb = load_notebook_node(f'{workflow_modules_path}/Phase-5-Diagnosing-Outputs-and-Generate-Report.ipynb')
    parameters['beast_xml_path'] = beast_xml_path
    parameters['report_template'] = report_template_path
    parameters['kernel_name'] = kernel_name
    diagnostic_nb['metadata']['kernelspec']['name'] = kernel_name
    diagnostic_nb['metadata']['kernelspec']['display_name'] = kernel_name
    diagnostic_nb = parameterize_notebook(diagnostic_nb, parameters, kernel_name)
    write_ipynb(diagnostic_nb, diagnostic_save_name)




import time
import os
from copy import deepcopy
import yaml
from papermill.inspection import _infer_parameters
from papermill.iorw import load_notebook_node
import numpy as np
import multiprocessing
import pandas as pd
from Bio import SeqIO
from datetime import datetime
from types import SimpleNamespace
from pandas.tseries.offsets import DateOffset
from beast_pype.path_utils import path_to_report_templates, path_to_workflow_modules, path_to_workflows

report_templates_path = path_to_report_templates()
workflow_modules_path = path_to_workflow_modules()
workflows_path = path_to_workflows()

def _gen_phase_4_params(
        save_dir,
        beast_seeds,
        beast_options_without_a_value=None,
        beast_options_needing_a_value=None,
        max_threads=None,
        sbatch_options_without_a_value=None,
        sbatch_options_needing_a_value=None):
    """
    Generate phase 4 parameters.

    Parameters
    ----------
    save_dir: str
        Path to directory where workflow outputs are saved.
    beast_seeds: list of ints or numpy.ndarray of ints
        Seeds used when running BEAST 2 in phase 4.
    beast_options_without_a_value: list of strs
        Single word arguments to pass to BEAST 2.
         For instance to use a GPU this would be ['-beagle_GPU'].
        See https://www.beast2.org/2021/03/31/command-line-options.html.
    beast_options_needing_a_value: dict
        Word followed by value arguments to pass to BEAST 2.
        For instance to use 3 threads when running BEAST 2 this would be: {'-threads': 3}.
        See https://www.beast2.org/2021/03/31/command-line-options.html.
    sbatch_options_without_a_value: list of strs
        Single word arguments to pass to sbatch.
        See https://slurm.schedmd.com/sbatch.html.
    sbatch_options_needing_a_value: dict
        Word followed by value arguments to pass to sbatch.
        See https://slurm.schedmd.com/sbatch.html.
    max_threads: int, optional
        Maximum number of threads to use.

    Returns
    -------
    Dictionary of phase 4 parameters.
    """
    if beast_options_without_a_value is None:
        beast_options_without_a_value = []
    if beast_options_needing_a_value is None:
        beast_options_needing_a_value = {}
    params_dict = {
        'save_dir': save_dir,
        'seeds': beast_seeds}
    if sbatch_options_without_a_value is not None or sbatch_options_needing_a_value is not None:
        params_dict['sbatch_arg_string'] = _to_arg_string(
            sbatch_options_without_a_value,
            sbatch_options_needing_a_value)
    else:
        if '-threads' in beast_options_needing_a_value:
            threads_per_run = beast_options_needing_a_value['-threads']
        else:
            threads_per_run = 1
        params_dict['jobs_per_time'] = int(max_threads / threads_per_run)
    params_dict['beast_arg_string'] = _to_arg_string(
            beast_options_without_a_value,
            beast_options_needing_a_value)
    return params_dict

class WorkflowParams(SimpleNamespace):
    """Base Class responsible for setting up and checking workflow parameters."""

    workflow_name = None
    not_to_be_recorded_as_params = ['sys', 'NamespaceMagics', 'get_ipython', 'debugpy', 'json', 'rng']


    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        kwargs = {key: value for key, value in kwargs.items() if key not in ['sys', 'NamespaceMagics', 'get_ipython', 'debugpy', 'json'] and not 'pydev' in key}
        super().__init__(**kwargs)
        # Check all the parameters are valid for this workflow.
        invalid_params = set(kwargs.keys()) - self.accepted_param_names()
        if invalid_params:
            raise ValueError(f'The following parameters that have been supplied to the {self.workflow_name} workflow are invalid:\n {", ".join(invalid_params)}')

        # Check for not being assigned
        if self.overall_save_dir is None:
            raise Exception(
                "overall_save_dir is missing from the parameters yml file, or\n" +
                "overall_save_dir has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for overall_save_dir."
            )

        if self.number_of_beast_runs is None:
            raise Exception(
                "number_of_beast_runs is missing from the parameters yml file, or\n" +
                "number_of_beast_runs has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for number_of_beast_runs."
            )
        if not isinstance(self.use_initial_tree, bool):
            raise TypeError('use_initial_tree must be True or False')
        if ' ' in self.overall_save_dir:
            raise ValueError('overall_save_dir cannot contain whitespace.')
        if self.fasta_path is not None and ' ' in self.fasta_path:
            raise ValueError('fasta_path cannot contain whitespace.')
        if self.log_file_basename is not None and ' ' in self.log_file_basename:
            raise ValueError('log_file_basename cannot contain whitespace.')
        if self.metadata_path is not None and not self.metadata_path.endswith(('.tsv', '.csv')):
            raise ValueError('Only .tsv and .csv are supported for metadata. Use correct file extension.')

        # Check values.
        if self.metadata_path is not None:
            if self.metadata_path.endswith('.tsv'):
                metadata_df = pd.read_csv(self.metadata_path, sep='\t')
            else:
                metadata_df = pd.read_csv(self.metadata_path)
            ids_in_metadata = metadata_df[self.sample_id_field]
            if ids_in_metadata.duplicated().any():
                duplicates_str = ', '.join(ids_in_metadata[ids_in_metadata.duplicated()].to_list())
                raise ValueError(f'There are duplicated values in the metadata field {self.sample_id_field}.\n.' +
                                 f' Check values {duplicates_str}.')
            if self.fasta_path is not None:
                sequence_lengths = [(record.id, len(record.seq)) for record in SeqIO.parse(self.fasta_path, 'fasta')]
                sequence_unique_lengths = set(record[1] for record in sequence_lengths)
                if len(sequence_unique_lengths) != 1:
                    raise ValueError(f'Fasta file {self.fasta_path} does not contain aligned fasta_path.')
                ids_in_fasta = pd.Series(record[0] for record in sequence_lengths)
                if ids_in_fasta.duplicated().any():
                    duplicates_str = ', '.join(ids_in_fasta[ids_in_fasta.duplicated()].to_list())
                    raise ValueError(f'There are duplicated id values in the fasta file.\n.' +
                                     f' Check values {duplicates_str}.')
                ids_in_metadata = set(ids_in_metadata)
                ids_in_fasta = set(ids_in_fasta)
                in_fasta_not_in_metadata = ids_in_fasta - ids_in_metadata
                in_metadata_not_in_fasta = ids_in_metadata - ids_in_fasta
                if len(in_fasta_not_in_metadata) > 0 or len(in_metadata_not_in_fasta) > 0:
                    in_fasta_not_in_metadata = ', '.join(in_fasta_not_in_metadata)
                    in_metadata_not_in_fasta = ', '.join(in_metadata_not_in_fasta)
                    raise ValueError(
                        f'There is a mismatch between the ids in the fasta at fasta_path and metadata field {self.sample_id_field}.\n' +
                        f"The following are in the metadata but not the fasta file: \n{in_metadata_not_in_fasta}.\n" +
                        f"The following are in the fasta file but not the metadata: \n{in_fasta_not_in_metadata}."
                    )
                try:
                    pd.to_datetime(metadata_df[self.collection_date_field], errors='raise')
                except:
                    raise ValueError(
                        f'The field {self.collection_date_field} in {self.metadata_path} cannot be read as a date.\n' +
                        f' Check the value in {self.collection_date_field}')
                self.number_of_sequences = len(ids_in_fasta)

        if self.root_strain_names is None and self.remove_root:
            raise ValueError('Root strain names must be specified if remove_root is True.')

        if self.max_threads is None:
            if 'SLURM_CPUS_PER_TASK' in os.environ:
                self.max_threads = int(os.environ['SLURM_CPUS_PER_TASK'])
            else:
                self.max_threads = multiprocessing.cpu_count() - 1

        # Checking MCMC options.
        if self.chain_length is not None:
            if not (isinstance(self.chain_length, int) and self.chain_length > 1):
                raise ValueError('If specified chain_length must be an integer greater than 1.')
        if self.trace_log_every is not None:
            if isinstance(self.trace_log_every, float) and self.trace_log_every < 1 and self.trace_log_every > 0:
                if not isinstance(self.chain_length, int):
                    raise AssertionError('If trace_log_every is a float, chain_length must be specified.')
                self.trace_log_every = int(round(self.trace_log_every*self.chain_length))
            elif not (isinstance(self.tree_log_every, int) and self.tree_log_every > 1):
                raise AssertionError('tree_log_every must be an integer greater than 1 or a float between 0 and 1.')
        if self.tree_log_every is not None:
            if isinstance(self.tree_log_every, float) and self.tree_log_every < 1 and self.tree_log_every > 0:
                if not isinstance(self.chain_length, int):
                    raise AssertionError('If tree_log_every is a float, chain_length must be specified.')
                self.tree_log_every = int(round(self.tree_log_every*self.chain_length))
            elif not (isinstance(self.tree_log_every, int) and self.tree_log_every > 1):
                raise AssertionError('tree_log_every must be an integer greater than 1 or a float between 0 and 1.')
        if self.screen_log_every is not None:
            if isinstance(self.screen_log_every, float) and self.screen_log_every < 1 and self.screen_log_every > 0:
                if not isinstance(self.chain_length, int):
                    raise AssertionError('If screen_log_every is a float, chain_length must be specified.')
                self.screen_log_every = int(round(self.screen_log_every*self.chain_length))
            elif not (isinstance(self.screen_log_every, int) and self.screen_log_every > 1):
                raise AssertionError('screen_log_every must be an integer greater than 1 or a float between 0 and 1.')
        if self.store_state_every is not None:
            if isinstance(self.store_state_every, float) and self.store_state_every < 1 and self.store_state_every > 0:
                if not isinstance(self.chain_length, int):
                    raise AssertionError('If store_state_every is a float, chain_length must be specified.')
                self.store_state_every = int(round(self.store_state_every*self.chain_length))
            elif not (isinstance(self.store_state_every, int) and self.store_state_every > 1):
                raise AssertionError('store_state_every must be an integer greater than 1 or a float between 0 and 1.')

        self.save_dir = self.overall_save_dir + '/' + self.specific_run_save_dir

        if self.metadata_path is not None:
            metadata_df.to_csv(f'{self.save_dir}/metadata.csv', index=False)

        if self.seed_to_rule_them_all is not None:
            if not isinstance(self.seed_to_rule_them_all, int) or self.seed_to_rule_them_all < 1:
                raise ValueError('seed_to_rule_them_all must be an integer greater than or equal to 1.')
            self.rng = np.random.default_rng(self.seed_to_rule_them_all)
            if self.iqtree_seed is not None:
                raise ValueError('If seed_to_rule_them_all is given, iqtree_seed should not be given as well.')
            else:
                self.iqtree_seed = self.rng.integers(low=0, high=int(9e6), size=1).tolist()[0]
            if self.treetime_seed is not None:
                raise ValueError('If seed_to_rule_them_all is given, treetime_seed should not be given as well.')
            else:
                self.treetime_seed = self.rng.integers(low=0, high=int(9e6), size=1).tolist()[0]
            if self.downsampled_iqtree_seed is not None:
                raise ValueError('If seed_to_rule_them_all is given, downsampled_iqtree_seed should not be given as well.')
            else:
                self.downsampled_iqtree_seed = self.rng.integers(low=0, high=int(9e6), size=1).tolist()[0]
            if self.downsampled_treetime_seed is not None:
                raise ValueError('If seed_to_rule_them_all is given, downsampled_treetime_seed should not be given as well.')
            else:
                self.downsampled_treetime_seed = self.rng.integers(low=0, high=int(9e6), size=1).tolist()[0]
            if self.beast_seeds is not None:
                raise ValueError('If seed_to_rule_them_all is given, beast_seeds should not be given as well.')
        else:
            self.rng = np.random.default_rng()
            if self.iqtree_seed is not None and (not isinstance(self.iqtree_seed, int) or self.iqtree_seed < 1):
                raise ValueError('iqtree_seed must be an integer greater than or equal to 1.')
            if self.treetime_seed is not None and (not isinstance(self.treetime_seed, int) or self.treetime_seed < 1):
                raise ValueError('treetime_seed must be an integer greater than or equal to 1.')
            if self.downsampled_iqtree_seed is not None and (not isinstance(self.downsampled_iqtree_seed, int) or self.downsampled_iqtree_seed < 1):
                raise ValueError('downsampled_iqtree_seed must be an integer greater than or equal to 1.')
            if self.downsampled_treetime_seed is not None and (not isinstance(self.downsampled_treetime_seed, int) or self.downsampled_treetime_seed < 1):
                raise ValueError('downsampled_treetime_seed must be an integer greater than or equal to 1.')

            if self.beast_seeds is not None:
                if isinstance(self.beast_seeds, int):
                    self.beast_seeds = [self.beast_seeds]
                if isinstance(self.beast_seeds, list):
                    if not all(isinstance(seed, int) and seed > 0 for seed in self.beast_seeds):
                        raise ValueError('If beast_seeds is a list, all values must be integers greater than or equal to 1.')
                else:
                    raise TypeError('beast_seeds must be a list of integers or a single integer.')

    def record_parameters(self, extras_dict={}):
        """
        Record the parameters used in workflow in `pipeline_run_info.yml`.

        Parameters
        ----------
        extras_dict: dict
            Dictionary of extra parameters to be recorded.
        """
        pipeline_run_info_path = f'{self.save_dir}/pipeline_run_info.yml'
        recordable_params = {key: value for key, value in self.__dict__.items() if key not in self.not_to_be_recorded_as_params}
        if os.path.exists(pipeline_run_info_path):
            with open(pipeline_run_info_path, 'r') as fp:
                to_record  = yaml.safe_load(fp)
            fp.close()
            if 'parameters' in to_record:
                to_record['parameters'].update(recordable_params)
            to_record.update(extras_dict)
        else:
            to_record = {'parameters': recordable_params, **extras_dict}
        with open(pipeline_run_info_path, 'w') as fp:
            yaml.dump(to_record, fp, sort_keys=True)
        fp.close()

    def retrieve_params(self, parameter_names):
        """
        Retrieve parameters held in an instance of this class.

        Parameters
        ----------
        parameter_names: list of strings
            List of parameter names to retrieve.

        Returns
        -------
        Dictionary of parameter names and values.
        """

        return {param_name: self.__dict__[param_name]  for param_name in parameter_names}

    def retrieve_phase_3_params(self, dir_to_check_and_save_to):
        """
        Retrieve parameters used in phase 3 of the workflow.

        Parameters
        ---------------
        dir_to_check_and_save_to: str
            Path to the directory to check for metadata, sequences and initial trees and
            save to.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        params = self.retrieve_params([
                                     'template_xml_path',
                                     'collection_date_field',
                                     'sample_id_field',
                                     'log_file_basename',
                                     'chain_length',
                                     'trace_log_every',
                                     'tree_log_every',
                                     'screen_log_every',
                                     'store_state_every'])
        params['save_dir'] = dir_to_check_and_save_to
        if os.path.exists(f'{dir_to_check_and_save_to}/downsampled_initial_tree/treetime.nwk'):
            params['initial_tree_path'] = f'{dir_to_check_and_save_to}/downsampled_initial_tree/treetime.nwk'
        elif os.path.exists(f'{dir_to_check_and_save_to}/initial_tree/treetime.nwk'):
            params['initial_tree_path'] = f'{dir_to_check_and_save_to}/initial_tree/treetime.nwk'
        elif os.path.exists(f'{dir_to_check_and_save_to}/initial_tree/treetime.nwk'):
            params['initial_tree_path'] = f'{dir_to_check_and_save_to}/initial_tree/iqtree.nwk'
        else:
            params['initial_tree_path'] = None

        if os.path.exists(f'{dir_to_check_and_save_to}/downsampled_metadata.csv'):
            params['metadata_path'] = f'{dir_to_check_and_save_to}/downsampled_metadata.csv'
        elif os.path.exists(f'{dir_to_check_and_save_to}/metadata.csv'):
            params['metadata_path'] = f'{dir_to_check_and_save_to}/metadata.csv'
        else:
            params['metadata_path'] = self.metadata_path
        if os.path.exists(f'{dir_to_check_and_save_to}/downsampled_sequences.fasta'):
            params['fasta_path'] = f'{dir_to_check_and_save_to}/downsampled_sequences.fasta'
        elif os.path.exists(f'{dir_to_check_and_save_to}/sequences.fasta'):
            params['fasta_path'] = f'{dir_to_check_and_save_to}/sequences.fasta'
        else:
            params['fasta_path'] = self.fasta_path
        return params

    def retrieve_phase_4_params(self):
        """
        Retrieve parameters used in phase 4 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        phase_4_params_dict = _gen_phase_4_params(**self.retrieve_params(['save_dir',
                                                                          'beast_seeds',
                                                                          'beast_options_without_a_value',
                                                                          'beast_options_needing_a_value',
                                                                          'max_threads',
                                                                          'sbatch_options_without_a_value',
                                                                          'sbatch_options_needing_a_value']))
        phase_4_params_dict.update(self.retrieve_params(['kernel_name']))
        return phase_4_params_dict

    def accepted_param_names(self):
        """
        Returns a set of accepted parameter names.

        Returns
        -------
        Set of strings.
        """
        nb = load_notebook_node(self.associated_workflow)
        parameters = _infer_parameters(nb)
        parameter_names = [parameter.name
                           for parameter in parameters]  + ['parent_workflow_parameters']
        return set(parameter_names)


class SimpleWorkflowParams(WorkflowParams):
    """Setup and check parameters for a simple workflows. """


    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)

        # Check for not being assigned
        if self.ready_to_go_xml is None:
            if self.fasta_path is None:
                raise Exception(
                    "A ready_to_go_xml has not being given and" +
                    "fasta_path is missing from the parameters yml file, or\n" +
                    "fasta_path has been given an 'null' value (which are converted to None in python)."
                )
            if self.template_xml_path is None:
                raise Exception(
                    "A ready_to_go_xml has not being given and" +
                    "template_xml_path is missing from the parameters yml file, or\n" +
                    "template_xml_path has been given an 'null' value (which are converted to None in python)."
                )
            if self.log_file_basename is None:
                raise Exception(
                    "A ready_to_go_xml has not being given and" +
                    "log_file_basename is missing from the parameters yml file, or\n" +
                    "log_file_basename has been given an 'null' value (which are converted to None in python)."
                )
            if self.metadata_path is None:
                raise Exception(
                    "A ready_to_go_xml has not being given and" +
                    "metadata_path is missing from the parameters yml file, or\n" +
                    "metadata_path has been given an 'null' value (which are converted to None in python).\n" +
                    "None/null values cannot be used for metadata_path."
                )
        else:
            if self.fasta_path is not None:
                raise Exception(
                    "A ready_to_go_xml has being given and:"
                    "fasta_path is in the parameters yml file, or\n" +
                    "fasta_path has not been given an 'null' value (which are converted to None in python)."
                )
            if self.template_xml_path is not None:
                raise Exception(
                    "A ready_to_go_xml has being given and:"
                    "template_xml_path is in the parameters yml file, or\n" +
                    "template_xml_path has not been given an 'null' value (which are converted to None in python)."
                )
            if self.log_file_basename is not None:
                raise Exception(
                    "A ready_to_go_xml has being given and:"
                    "log_file_basename is in the parameters yml file, or\n" +
                    "log_file_basename has not been given an 'null' value (which are converted to None in python)."
                )

        if self.ready_to_go_xml is not None or not self.use_initial_tree:  # None of the parameters below are needed if a ready_to_go_xml is provided.
            if self.initial_tree_type not in ['Temporal', 'Distance', None]:
                raise ValueError('initial_tree_type must be either "Temporal" or "Distance" or None.')

            if self.initial_tree_path is not None and self.initial_tree_type is None:
                raise ValueError('initial_tree_type must be specified if initial_tree_path is given.')

            if self.downsample_to is not None and (self.initial_tree_path is not None or self.initial_tree_type == 'Distance'):
                raise ValueError("Currently beast_pype's downsample_toing method is tied to its Tree Time tree building." +
                                 "Therefore, to use this down sampling method an initial_tree_path should not be given and an initial_tree_type should be set to 'Temporal'.")

        number_of_seeds = self.number_of_beast_runs
        if self.beast_seeds is not None and len(self.beast_seeds) != number_of_seeds:
            raise ValueError('If beast_seeds is given, the number of seeds provided should match the number of BEAST runs to be performed (number_of_beast_runs).')
        # If beast_seeds not given generate them.
        # BEASTs random number seed can select the same seed for multiple runs if they are launched close together in time (such as programmatically). Therefore, numpy is used to generate beast_seeds for running BEAST.
        if self.beast_seeds is None:
            self.beast_seeds = self.rng.integers(low=0, high=int(9e5), size=number_of_seeds).tolist()

    def retrieve_phase_2i_params(self):
        """
        Retrieve parameters used in phase 2i of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.ready_to_go_xml is None and \
                self.use_initial_tree and \
                (self.initial_tree_path is None):
            params = self.retrieve_params(['save_dir',
                                           'fasta_path' ,
                                           'max_threads',
                                           'kernel_name'])
            params['tree_dir_name'] = 'initial_tree'
            if self.iqtree_seed is not None:
                params['seed'] = self.iqtree_seed
        else:
            params = None
        return params
    
    def retrieve_phase_2i_correction_params(self):
        """
        Retrieve parameters used in the correction of phase 2i of the workflow.

        Parameters
        -------------
        xml_set_directory: str
            Path to the XML set directory to include under 'save_dir' of the returned
            dictionary.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        correction_params = self.retrieve_params(['save_dir', 'fasta_path'])
        correction_params['tree_dir_name'] = 'initial_tree'
        return correction_params

    def retrieve_phase_2ii_params(self):
        """
        Retrieve parameters used in phase 2ii of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.ready_to_go_xml is None and \
                self.use_initial_tree and \
                (self.initial_tree_path is None and self.initial_tree_type == 'Temporal'):
            params = self.retrieve_params(['save_dir',
                                           'fasta_path',
                                           'metadata_path',
                                           'sample_id_field',
                                           'collection_date_field',
                                           'downsample_to',
                                           'root_strain_names',
                                           'remove_root'])
            params['tree_dir_name'] = 'initial_tree'
            if self.treetime_seed is not None:
                params['seed'] = self.treetime_seed
        else:
            params = None
        return params

    def retrieve_phase_2iii_params(self):
        """
        Retrieve parameters used in phase 2i of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if os.path.exists(f'{self.save_dir}/downsampled_sequences.fasta'):
            params = self.retrieve_params(['save_dir', 'max_threads', 'kernel_name'])
            params['fasta_path'] = f'{self.save_dir}/downsampled_sequences.fasta'
            params['tree_dir_name'] = 'downsampled_initial_tree'
            if self.downsampled_iqtree_seed is not None:
                params['seed'] = self.downsampled_iqtree_seed
        else:
            params = None
        return params
    
    def retrieve_phase_2iii_correction_params(self):
        """
        Retrieve parameters used in the correction of phase 2iii of the workflow.

        Parameters
        -------------
        xml_set_directory: str
            Path to the XML set directory to include under 'save_dir' of the returned
            dictionary.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        params = self.retrieve_params(['save_dir'])
        params['fasta_path'] = f'{self.save_dir}/downsampled_sequences.fasta'
        params['tree_dir_name'] = 'downsampled_initial_tree'
        return params

    def retrieve_phase_2iv_params(self):
        """
        Retrieve parameters used in phase 2iv of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if os.path.exists(f'{self.save_dir}/downsampled_sequences.fasta'):
            params = self.retrieve_params(['save_dir',
                                           'sample_id_field',
                                           'collection_date_field'])
            params['fasta_path'] = f'{self.save_dir}/downsampled_sequences.fasta'
            params['metadata_path'] = f'{self.save_dir}/downsampled_metadata.csv'
            params['tree_dir_name'] = 'downsampled_initial_tree'
            if self.downsampled_treetime_seed is not None:
                params['seed'] = self.downsampled_treetime_seed
        else:
            params = None
        return params

    def retrieve_phase_5_params(self):
        """
        Retrieve parameters used in phase 5 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        phase_5_params_dict = self.retrieve_params(['kernel_name'])
        if self.report_template is None:
            phase_5_params_dict['report_template'] = self.default_report_template
        else:
            phase_5_params_dict['report_template'] = self.report_template
        return phase_5_params_dict

    def tip_date_range(self):
        """
        Generate tip date range.
        Returns
        -------
        oldest_tip_date: datetime
            Oldest tip date.
        youngest_tip_date: datetime
            Youngest tip date.

        """
        if self.downsample_to is None:
            metadata_path = self.metadata_path
        else:
            metadata_path = f'{self.save_dir}/downsampled_metadata.csv'
        if metadata_path.endswith('.tsv'):
            sep = '\t'
        elif metadata_path.endswith('.csv'):
            sep = ','
        else:
            raise ValueError('Only .tsv and .csv are supported for metadata. Use correct file extension.')
        metadata = pd.read_csv(metadata_path, parse_dates=[self.collection_date_field], sep=sep)
        youngest_tip_date = metadata[self.collection_date_field].max()
        oldest_tip_date = metadata[self.collection_date_field].min()
        return oldest_tip_date, youngest_tip_date


class GenericWorkflowParams(SimpleWorkflowParams):
    """Set up and check Generic workflow parameters."""


    workflow_name = 'Generic'
    associated_workflow = f"{workflows_path}/Generic.ipynb"
    default_report_template = "Generic"

    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)
        if self.ready_to_go_xml is not None and self.metadata_path is not None:
                raise Exception(
                    "A ready_to_go_xml has being given and"
                    "metadata_path is in the parameters yml file, or\n" +
                    "metadata_path has not been given an 'null' value (which are converted to None in python)."
                )
        self.record_parameters()
        self.sampling_prop_partition_dates = None
        self.rt_partition_dates = None

    def retrieve_phase_3_params(self):
        """
        Retrieve parameters used in phase 3 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.ready_to_go_xml is None:
            params = super().retrieve_phase_3_params(dir_to_check_and_save_to=self.save_dir)
        else:
            params = None
        return params

def _bdksy_serial_errors(rt_dims,
                         rt_partitions,
                         sampling_prop_dims,
                         sampling_prop_partitions,
                         zero_sampling_before_first_sample,
                         ready_to_go_xml,
                         use_initial_tree,
                         origin_start_addition,
                         initial_tree_type,
                         origin_upper_addition):
    if rt_dims is not None and rt_partitions is not None:
        raise TypeError('rt_dims and rt_partitions cannot be used together')
    if sampling_prop_dims is not None and sampling_prop_partitions is not None:
        raise TypeError('sampling_prop_dims and sampling_prop_partitions cannot be used together')
    if sampling_prop_dims is not None and zero_sampling_before_first_sample:
        raise TypeError('Currently zero_sampling_before_first_sample can only be used '+
                        'on its own or with sampling_prop_partitions, but NOT with sampling_prop_dims.')
    if ready_to_go_xml is not None or not use_initial_tree:
        if origin_start_addition is not None:
            if initial_tree_type != 'Temporal':
                raise ValueError('origin_start_addition is reliant on the initial_tree_type being "Temporal".')
            if origin_upper_addition is None:
                raise ValueError('origin_start_addition is reliant on origin_upper_addition being given as well.')

        if origin_upper_addition is not None:
            if initial_tree_type != 'Temporal':
                raise ValueError('origin_upper_addition is reliant on the initial_tree_type being "Temporal".')
            if origin_upper_addition is None:
                raise ValueError('origin_upper_addition is reliant on origin_start_addition being given as well.')

def _partition_dates_dict_to_list(change_dates_dict, youngest_tip_date, oldest_tip_date, parameter_name, inclusive_of_end=False):
    if 'end' in change_dates_dict:
        end = change_dates_dict['end']
        if isinstance(end, str):
            end = datetime.strptime(end, '%Y-%m-%d')
        if not isinstance(end, datetime):
            raise TypeError(f'{parameter_name}["end"] should be a datetime object or string of format YYYY-MM-DD')
    else:
        end = oldest_tip_date
    change_dates_list = []
    offsets = {change_dates_dict['unit']: change_dates_dict['every']}
    date_to_append = youngest_tip_date - DateOffset(**offsets)
    if not date_to_append > end:
        raise ValueError(
            f'Change dates dictionary for {parameter_name} makes no sense.\n' +
            'First partition date generated is not after the "end" date.\n' +
            'Suggest using smaller "units" or "every" values or pushing "end" date back.'
                         )
    while date_to_append > end:
        change_dates_list.append(date_to_append.strftime('%Y-%m-%d'))
        date_to_append = date_to_append - DateOffset(**offsets)
    if inclusive_of_end:
        change_dates_list.append(end.strftime('%Y-%m-%d'))
    return change_dates_list

def _phase_3_partition_params(rt_partitions,
                              sampling_prop_partitions,
                              sampling_prop_include_oldest_tip_date,
                              youngest_tip_date,
                              oldest_tip_date):
    if isinstance(rt_partitions, dict):
        rt_partition_dates = _partition_dates_dict_to_list(rt_partitions,
                                                        youngest_tip_date,
                                                        oldest_tip_date,
                                                        parameter_name='rt_partitions')
    else:
        rt_partition_dates = rt_partitions
    if isinstance(sampling_prop_partitions, dict):
        sampling_prop_partition_dates = _partition_dates_dict_to_list(sampling_prop_partitions,
                                                                   youngest_tip_date,
                                                                   oldest_tip_date,
                                                                   parameter_name='sampling_prop_partitions',
                                                                   inclusive_of_end=sampling_prop_include_oldest_tip_date
                                                                   )
    else:
        sampling_prop_partition_dates = sampling_prop_partitions
    return rt_partition_dates, sampling_prop_partition_dates

_accepted_bdsky_param = {'origin_start_addition',
                         'origin_upper_addition',
                         'origin_prior',
                         'rt_dims',
                         'rt_partitions',
                         'sampling_prop_dims',
                         'sampling_prop_partitions',
                         'zero_sampling_before_first_sample'
                         }



class BDSKYSerialWorkflowParams(SimpleWorkflowParams):
    """Set up and check BDSKY-Serial workflow parameters."""

    workflow_name = 'BDSKY-Serial'
    associated_workflow = f"{workflows_path}/BDSKY-Serial.ipynb"
    default_report_template = "BDSKY-Serial"

    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)
        _bdksy_serial_errors(
            rt_dims=self.rt_dims,
            rt_partitions=self.rt_partitions,
            sampling_prop_dims=self.sampling_prop_dims,
            sampling_prop_partitions=self.sampling_prop_partitions,
            zero_sampling_before_first_sample=self.zero_sampling_before_first_sample,
            ready_to_go_xml=self.ready_to_go_xml,
            use_initial_tree=self.use_initial_tree,
            origin_start_addition=self.origin_start_addition,
            initial_tree_type=self.initial_tree_type,
            origin_upper_addition=self.origin_upper_addition
        )
        self.record_parameters()
        self.sampling_prop_partition_dates = None
        self.rt_partition_dates = None

    def retrieve_phase_3_params(self):
        """
        Retrieve parameters used in phase 3 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.ready_to_go_xml is None:
            phase_3_params = super().retrieve_phase_3_params(dir_to_check_and_save_to=self.save_dir)
            phase_3_params.update(self.retrieve_params([
                                                   'rt_dims',
                                                   'sampling_prop_dims',
                                                   'zero_sampling_before_first_sample',
                                                   'origin_start_addition',
                                                   'origin_upper_addition',
                                                   'origin_prior']))
            if isinstance(self.rt_partitions, dict) or isinstance(self.sampling_prop_partitions, dict):
                oldest_tip_date, youngest_tip_date = self.tip_date_range()
                rt_partition_dates, sampling_prop_partition_dates = _phase_3_partition_params(
                    self.rt_partitions,
                    self.sampling_prop_partitions,
                    self.zero_sampling_before_first_sample,
                    youngest_tip_date,
                    oldest_tip_date)
                phase_3_params['rt_partitions'] = rt_partition_dates
                phase_3_params['sampling_prop_partitions'] = sampling_prop_partition_dates
            else:
                phase_3_params['rt_partitions'] = self.rt_partitions
                phase_3_params['sampling_prop_partitions'] = self.sampling_prop_partitions
                if self.zero_sampling_before_first_sample:
                    oldest_tip_date, youngest_tip_date = self.tip_date_range()
                    if self.sampling_prop_partitions is not None:
                        partition_dates = pd.to_datetime(self.sampling_prop_partitions)
                        if oldest_tip_date > min(partition_dates):
                            raise ValueError('If using zero_sampling_before_first_sample oldest partition date should be before oldest date in the list from of sampling_prop_partitions.')
                        phase_3_params['sampling_prop_partitions'] = [oldest_tip_date.strftime('%Y-%m-%d')] + phase_3_params['sampling_prop_partitions']
                    else:
                        phase_3_params['sampling_prop_partitions'] = [oldest_tip_date.strftime('%Y-%m-%d')]

            with open(f'{self.save_dir}/pipeline_run_info.yml', 'r') as file:
                pipeline_run_info = yaml.safe_load(file)
            self.rt_partition_dates = phase_3_params['rt_partitions']
            pipeline_run_info['rt_partitions'] = phase_3_params['rt_partitions']
            self.sampling_prop_partition_dates = phase_3_params['sampling_prop_partitions']
            pipeline_run_info['sampling_prop_partitions'] = phase_3_params['sampling_prop_partitions']
            with open(f'{self.save_dir}/pipeline_run_info.yml', 'w') as fp:
                yaml.dump(pipeline_run_info, fp, sort_keys=True)
            fp.close()
        else:
            phase_3_params = None
        return phase_3_params


class ComparativeWorkflowParams(WorkflowParams):
    """Setup and check parameters for comparative workflows. """

    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)
        #Check for not being assigned
        if self.fasta_path is None:
            raise Exception(
                "A ready_to_go_xml has not being given and" +
                "fasta_path is missing from the parameters yml file, or\n" +
                "fasta_path has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for fasta_path."
            )
        if self.template_xml_path is None:
            raise Exception(
                "template_xml_path is missing from the parameters yml file, or\n" +
                "template_xml_path has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for template_xml_path."
            )
        if self.log_file_basename is None:
            raise Exception(
                "log_file_basename is missing from the parameters yml file, or\n" +
                "log_file_basename has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for log_file_basename."
            )
        if self.metadata_path is None:
            raise Exception(
                "metadata_path is missing from the parameters yml file, or\n" +
                "metadata_path has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for metadata_path."
            )
        if self.xml_set_definitions is None:
            raise Exception(
                "xml_set_definitions is missing from the parameters yml file, or\n" +
                "xml_set_definitions has been given an 'null' value (which are converted to None in python).\n" +
                "None/null values cannot be used for xml_set_definitions."
            )
        if not self.use_initial_tree:  # None of the parameters below are needed if you are not using an inial tree.
            if self.initial_tree_type not in ['Temporal', 'Distance']:
                raise ValueError('initial_tree_type must be either "Temporal" or "Distance".')

            if self.downsample_to is not None and (self.initial_tree_type != 'Temporal'):
                raise ValueError("Currently beast_pype's downsample_toing method is tied to its Tree Time tree building." +
                                 "Therefore, to use this down sampling method an use_initial_tree = True and initial_tree_type = 'Temporal'.")

        number_of_seeds = self.number_of_beast_runs * len(self.xml_set_definitions)
        if self.beast_seeds is not None and len(self.beast_seeds) != number_of_seeds:
            raise ValueError('If beast_seeds is given, the number of seeds provided should match the number of BEAST runs to be performed (number_of_beast_runs multiplied by the number of XML sets defined in xml_set_definitions).')
        # If beast_seeds not given generate them.
        # BEASTs random number seed can select the same seed for multiple runs if they are launched close together in time (such as programmatically). Therefore, numpy is used to generate beast_seeds for running BEAST.
        if self.beast_seeds is None:
            self.beast_seeds = self.rng.integers(low=0, high=int(9e5), size=number_of_seeds).tolist()

    def gen_xml_set_directories(self):
        """
        Generate the XML set directories for use in the workflow.
        """
        xml_set_directories = {}
        for xml_set in self.xml_set_definitions.keys():
            xml_set_directory = f'{os.getcwd()}/{self.save_dir}/{xml_set}'
            os.makedirs(xml_set_directory)
            xml_set_directories[xml_set] = xml_set_directory

        return xml_set_directories


    def retrieve_phase_1_params(self):
        """
        Retrieve parameters used in phase 1 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        return self.retrieve_params(['save_dir',
                                     'xml_set_definitions',
                                     'xml_set_directories',
                                     'metadata_path',
                                     'data_filter',
                                     'sample_id_field',
                                     'collection_date_field',
                                     'fasta_path'])

    def retrieve_phase_2i_params(self):
        """
        Retrieve parameters used in phase 2i of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.use_initial_tree:
            params = self.retrieve_params(['save_dir', 'max_threads', 'kernel_name'])
            params['fasta_path'] = 'sequences.fasta'
            params['tree_dir_name'] = 'initial_tree'
            if self.iqtree_seed is not None:
                params['seed'] = self.iqtree_seed
        else:
            params = None
        return params
    
    def retrieve_phase_2i_correction_params(self, xml_set_directory):
        """
        Retrieve parameters used in the correction of phase 2i of the workflow.

        Parameters
        -------------
        xml_set_directory: str
            Path to the XML set directory to include under 'save_dir' of the returned
            dictionary.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        parameters = {
            'save_dir': xml_set_directory,
            'tree_dir_name': 'initial_tree',
            'fasta_path': f'{xml_set_directory}/sequences.fasta'}
        return parameters

    def retrieve_phase_2ii_params(self, xml_set_directory):
        """
        Retrieve parameters used in phase 2ii of the workflow.

        Parameters
        -------------
        xml_set_directory: str
            Path to the XML set directory to include under 'save_dir' of the returned
            dictionary.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.use_initial_tree and self.initial_tree_type == 'Temporal':
            parameters = self.retrieve_params([
                'sample_id_field',
                'collection_date_field',
                'downsample_to',
                'root_strain_names',
                'remove_root'])
            parameters['save_dir'] = xml_set_directory
            parameters['tree_dir_name'] = 'initial_tree'
            if self.treetime_seed is not None:
                parameters['seed'] = self.treetime_seed
        else:
            parameters = None
        return parameters

    def retrieve_phase_2iii_params(self):
        """
        Retrieve parameters used in phase 2i of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.downsample_to is not None:
            params = self.retrieve_params(['save_dir', 'max_threads', 'kernel_name'])
            params['fasta_path'] = 'downsampled_sequences.fasta'
            params['tree_dir_name'] = 'downsampled_initial_tree'
            if self.downsampled_iqtree_seed is not None:
                params['seed'] = self.downsampled_iqtree_seed
        else:
            params = None
        return params
    
    def retrieve_phase_2iii_correction_params(self, xml_set_directory):
        """
        Retrieve parameters used in the correction of phase 2iii of the workflow.

        Parameters
        -------------
        xml_set_directory: str
            Path to the XML set directory to include under 'save_dir' of the returned
            dictionary.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if os.path.exists(f'{xml_set_directory}/downsampled_sequences.fasta'):
            parameters = {
                'save_dir': xml_set_directory,
                'tree_dir_name': 'downsampled_initial_tree',
                'fasta_path': f'{xml_set_directory}/downsampled_sequences.fasta'}
        else:
            parameters = None
        return parameters


    def retrieve_phase_2iv_params(self, xml_set_directory):
        """
        Retrieve parameters used in phase 2iv of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if os.path.exists(f'{xml_set_directory}/downsampled_sequences.fasta'):
            params = self.retrieve_params([
                                           'sample_id_field',
                                           'collection_date_field'])
            params['save_dir'] = xml_set_directory
            params['fasta_path'] = f'{xml_set_directory}/downsampled_sequences.fasta'
            params['metadata_path'] = f'{xml_set_directory}/downsampled_metadata.csv'
            params['tree_dir_name'] = 'downsampled_initial_tree'
            if self.downsampled_treetime_seed is not None:
                params['seed'] = self.downsampled_treetime_seed
        else:
            params = None
        return params

    def retrieve_phase_5_params(self):
        """
        Retrieve parameters used in phase 5 of the workflow.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.xml_set_label is not None:
            phase_5_params_dict = self.retrieve_params([
                'xml_set_label', 'kernel_name'])
        else:
            phase_5_params_dict = {'xml_set_label': 'xml set',
                                   **self.retrieve_params(['kernel_name'])}
        if self.report_template is None:
            phase_5_params_dict['report_template'] = self.default_report_template
        else:
            phase_5_params_dict['report_template'] = self.report_template
        return phase_5_params_dict

    def tip_date_range(self, xml_set_directory):
        """
        Generate tip date range.

        Parameters
        -------------
        xml_set_directory: str
            Path to XML set directory where metadata is held.

        Returns
        -------
        oldest_tip_date: datetime
            Oldest tip date.
        youngest_tip_date: datetime
            Youngest tip date.
        """
        metadata_path = f'{xml_set_directory}/downsampled_metadata.csv'
        if not os.path.exists(metadata_path):
            metadata_path = f'{xml_set_directory}/metadata.csv'
            
        if metadata_path.endswith('.tsv'):
            sep = '\t'
        elif metadata_path.endswith('.csv'):
            sep = ','
        else:
            raise ValueError('Only .tsv and .csv are supported for metadata. Use correct file extension.')
        metadata = pd.read_csv(metadata_path, parse_dates=[self.collection_date_field], sep=sep)
        youngest_tip_date = metadata[self.collection_date_field].max()
        oldest_tip_date = metadata[self.collection_date_field].min()
        return oldest_tip_date, youngest_tip_date


class GenericComparativeWorkflowParams(ComparativeWorkflowParams):
    """Set up and check Generic Comparative workflow parameters."""

    workflow_name = 'Generic Comparative'
    associated_workflow = f"{workflows_path}/Generic-Comparative.ipynb"
    default_report_template = "Generic-Comparative"

    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)
        xml_set_directories = self.gen_xml_set_directories()
        self.record_parameters(extras_dict={'xml set directories': xml_set_directories})
        self.sampling_prop_partition_dates = None
        self.rt_partition_dates = None
        self.xml_set_directories = xml_set_directories

    def retrieve_phase_3_params(self, xml_set_directory):
        phase_3_params = super().retrieve_phase_3_params(dir_to_check_and_save_to=xml_set_directory)
        return phase_3_params

class BDSKYSerialComparativeWorkflowParams(ComparativeWorkflowParams):
    """Set up and check BDSKY-Serial Comparative workflow parameters."""

    workflow_name = 'BDSKY-Serial Comparative'
    associated_workflow = f"{workflows_path}/BDSKY-Serial-Comparative.ipynb"
    default_report_template = "BDSKY-Serial-Comparative"

    def __init__(self, **kwargs):
        """
        Initialise the workflow parameters (perform checks and set up).

        Parameters
        ----------
        kwargs:
            Parameters to use in the workflow.
        """
        super().__init__(**kwargs)
        _bdksy_serial_errors(
            rt_dims=self.rt_dims,
            rt_partitions=self.rt_partitions,
            sampling_prop_dims=self.sampling_prop_dims,
            sampling_prop_partitions=self.sampling_prop_partitions,
            zero_sampling_before_first_sample=self.zero_sampling_before_first_sample,
            ready_to_go_xml=None,
            use_initial_tree=self.use_initial_tree,
            origin_start_addition=self.origin_start_addition,
            initial_tree_type=self.initial_tree_type,
            origin_upper_addition=self.origin_upper_addition
        )
        xml_set_directories = self.gen_xml_set_directories()
        self.record_parameters(extras_dict={'xml set directories': xml_set_directories})
        self.xml_set_directories = xml_set_directories
        self.sampling_prop_partition_dates = None
        self.rt_partition_dates = None

    def retrieve_phase_3_params(self, xml_set, xml_set_directory):
        """
        Retrieve parameters used in phase 3 of the workflow.

        Parameters
        ----------
        xml_set: str
            Which xml_set are phase 3 parameters being retrieved for.
        xml_set_directory: str
            Path to XML set directory.

        Returns
        -------
        Dictionary of parameter names and values.
        """
        if self.rt_partitions is not None and xml_set in self.rt_partitions:
            rt_partitions_to_use = self.rt_partitions[xml_set]
        else:
            rt_partitions_to_use = self.rt_partitions

        if self.sampling_prop_partitions is not None and xml_set in self.sampling_prop_partitions:
            sampling_prop_partitions_to_use = self.sampling_prop_partitions[xml_set]
        else:
            sampling_prop_partitions_to_use = self.sampling_prop_partitions
        phase_3_params = super().retrieve_phase_3_params(dir_to_check_and_save_to=xml_set_directory)
        phase_3_params.update(self.retrieve_params([
            'rt_dims',
            'sampling_prop_dims',
            'zero_sampling_before_first_sample',
            'origin_start_addition',
            'origin_upper_addition',
            'origin_prior']))
        if isinstance(rt_partitions_to_use, dict) or isinstance(sampling_prop_partitions_to_use, dict):
            oldest_tip_date, youngest_tip_date = self.tip_date_range(xml_set_directory)
            rt_partition_dates, sampling_prop_partition_dates = _phase_3_partition_params(
                rt_partitions_to_use,
                sampling_prop_partitions_to_use,
                self.zero_sampling_before_first_sample,
                youngest_tip_date,
                oldest_tip_date)
            phase_3_params['rt_partitions'] = rt_partition_dates
            phase_3_params['sampling_prop_partitions'] = sampling_prop_partition_dates
        else:
            phase_3_params['rt_partitions'] = rt_partitions_to_use
            phase_3_params['sampling_prop_partitions'] = sampling_prop_partitions_to_use
            if self.zero_sampling_before_first_sample:
                oldest_tip_date, youngest_tip_date = self.tip_date_range(xml_set_directory)
                if sampling_prop_partitions_to_use is not None:
                    partition_dates = pd.to_datetime(sampling_prop_partitions_to_use)
                    if oldest_tip_date > min(partition_dates):
                        raise ValueError('If using zero_sampling_before_first_sample oldest partition date should be before oldest date in the list from of sampling_prop_partitions.')
                    phase_3_params['sampling_prop_partitions'] = [oldest_tip_date.strftime('%Y-%m-%d')] + phase_3_params['sampling_prop_partitions']
                else:
                    phase_3_params['sampling_prop_partitions'] = [oldest_tip_date.strftime('%Y-%m-%d')]

        with open(f'{self.save_dir}/pipeline_run_info.yml', 'r') as file:
            pipeline_run_info = yaml.safe_load(file)

        if 'rt_partitions' not in pipeline_run_info:
            self.rt_partition_dates = {xml_set: phase_3_params['rt_partitions']}
            pipeline_run_info['rt_partitions'] = {xml_set: phase_3_params['rt_partitions']}
        else:
            self.rt_partition_dates[xml_set] = phase_3_params['rt_partitions']
            pipeline_run_info['rt_partitions'][xml_set] = phase_3_params['rt_partitions']

        if 'sampling_prop_partitions' not in pipeline_run_info:
            self.sampling_prop_partition_dates = {
                xml_set: phase_3_params['sampling_prop_partitions']}
            pipeline_run_info['sampling_prop_partitions'] = {
                xml_set: phase_3_params['sampling_prop_partitions']}
        else:
            self.sampling_prop_partition_dates[xml_set] = phase_3_params[
                'sampling_prop_partitions']
            pipeline_run_info['sampling_prop_partitions'][xml_set] = phase_3_params[
                'sampling_prop_partitions']

        with open(f'{self.save_dir}/pipeline_run_info.yml', 'w') as fp:
            yaml.dump(pipeline_run_info, fp, sort_keys=True)
        fp.close()
        return phase_3_params


def check_file_for_phrase(file_path,
                          phrase='slurm_job_complete',
                          wait_time=60):
    """
    If file exists check file for presence of phrase. Not found wait_time in seconds.

    Used to check a slurm .out file if the slurm job is complete,
    provided slurm sbatch ends with `; echo slurm_job_complete`.

    Parameters
    ----------
    file_path: str
       Path to text file (slurm .out file).
    phrase: str (default='slurm_job_complete')
       Phrase to check for file present at file_path.
    wait_time : int (default=60)
       Time in seconds to wait between each check.

    Returns
    -------
    None

    """
    # Section is silenced as the slurm job is not started immediately.
    #if not os.path.exists(file_path):
    #    raise FileNotFoundError(f'file_path "{file_path}" does not exist.')
    #if not os.path.isfile(file_path):
    #    raise FileNotFoundError(f'file_path "{file_path}" is not a file.')
        
    complete_phrase_found = False
    while not complete_phrase_found:
        if os.path.isfile(file_path):
            with open(file_path) as file:
                s = file.read()
                complete_phrase_found = phrase in s
            file.close()
        if not complete_phrase_found:
            time.sleep(wait_time)

def _to_arg_string(options_without_a_value, options_needing_a_value):
    if options_without_a_value is None:
        list_all_args = []
    else:
        list_all_args = deepcopy(options_without_a_value)
    if options_needing_a_value is None:
        options_needing_a_value = {}
    list_all_args += [key + ' ' + str(value) for key, value in options_needing_a_value.items()]
    return ' '.join(list_all_args)

def setup_optimising_config(name,
                            configuration,
                            save_dir,
                            ready_to_go_xml,
                            threads_arg=None,
                            instances_arg=None,
                            cpu_arg=None
                            ):
    """
    Set up a configuration to test in Optimising-BEAST-Runs workflow.

    Parameters
    ----------
    name: str
    configuration: dict
    save_dir: str
    ready_to_go_xml: str
        Path of ready to go xml file.
    threads_arg: int
    instances_arg: int
    cpu_arg: int

    Returns
    -------
    Dictionary of parameter names and values.
    """
    configuration = deepcopy(configuration)
    save_name = deepcopy(name)
    if threads_arg is not None:
        if 'beast_options_needing_a_value' not in configuration:
            configuration['beast_options_needing_a_value'] = {}
        if '-threads' in configuration['beast_options_needing_a_value']:
            raise ValueError('thread_options should not be being if -threads inside is in beast_options_needing_a_value.')
        configuration['beast_options_needing_a_value']['-threads']=threads_arg
        save_name = f'{save_name}_threads_{str(threads_arg)}'
    if instances_arg is not None:
        if 'beast_options_needing_a_value' not in configuration:
            configuration['beast_options_needing_a_value'] = {}
        if '-instances' in configuration['beast_options_needing_a_value']:
            raise ValueError('thread_options should not be being if -instances inside is in beast_options_needing_a_value.')
        configuration['beast_options_needing_a_value']['-instances']=instances_arg
        save_name = f'{save_name}_instances_{str(instances_arg)}'
    if cpu_arg is not None:
        if 'sbatch_options_needing_a_value' not in configuration:
            configuration['sbatch_options_needing_a_value'] = {}
        if '--cpus-per-task' in configuration['sbatch_options_needing_a_value']:
            raise ValueError('cpu_options should not be being if --cpus-per-task inside is in sbatch_options_needing_a_value.')
        configuration['sbatch_options_needing_a_value']['--cpus-per-task']=cpu_arg
        save_name = f'{save_name}_cpu_{str(cpu_arg)}'
    if 'sbatch_options_needing_a_value' in configuration and '--job-name' not in configuration['sbatch_options_needing_a_value']:
        configuration['sbatch_options_needing_a_value']['--job-name'] = save_name
    if not ('sbatch_options_needing_a_value' in configuration or 'sbatch_single_wor_args'  in configuration):
        if 'max_threads' not in configuration:
            configuration['max_threads'] = multiprocessing.cpu_count() - 1


    if 'beast_seeds' not in configuration:
        configuration['beast_seeds'] = np.random.randint(low=1, high=int(1e6), size=configuration['number_of_beast_runs']).tolist()
    del configuration['number_of_beast_runs']
    save_dir = f'{save_dir}/{save_name}'
    os.makedirs(save_dir)
    parameters = _gen_phase_4_params(
        save_dir=save_dir,
        **configuration
    )

    parameters['beast_xml'] = ready_to_go_xml
    with open(f'{save_dir}/configuration_run_info.yml', 'w') as fp:
        yaml.dump(parameters, fp, sort_keys=True)
    fp.close()
    return parameters


from base_test_classes import ComparativeWorkflowVariationTest

xml_generation_notebook = 'Phase-3-Gen-BDSKY-Serial-xml.ipynb'
workflow = 'COVID-Strain-Surveillance'

class TestCOVIDStrainSurveillance(ComparativeWorkflowVariationTest):
    name = 'Test_COVID_Strain_Surveillance'
    parameters_path = 'tests/parameters/COVID-Strain-Surveillance.yml'
    workflow = workflow
    initial_tree_building = True
    downsampling = False
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

    @property
    def xml_set_labels(self):
        return [f'VOI_{strain}' for strain in self.parameters['voi_strains']] + [f"DR_{self.parameters['dr_strain']}"]
